###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCbDIRAC.AccountingSystem.private.Plotters.StoragePlotter.

StoragePlotter.__bases__:
  DIRAC.AccountingSystem.private.Plotters.BaseReporter.BaseReporter
"""
from DIRAC import S_OK, S_ERROR
from DIRAC.AccountingSystem.private.Plotters.BaseReporter import BaseReporter

from LHCbDIRAC.AccountingSystem.Client.Types.Storage import Storage

# FIXME: refactor _reportMethods
# FIXME: refactor _plotMethods


class StoragePlotter(BaseReporter):
    """StoragePlotter as extension of BaseReporter."""

    _typeName = "Storage"
    _typeKeyFields = [dF[0] for dF in Storage().definitionKeyFields]

    # .............................................................................
    # catalog Space

    _reportCatalogSpaceName = "LFN size"

    def _reportCatalogSpace(self, reportRequest):
        """Reports about LFN size and catalog space from the accounting.

        :param reportRequest: <dict>
          { 'grouping'       : 'Directory',
            'groupingFields' : ( '%s', [ 'Directory' ] ),
            'startTime'      : 1355663249.0,
            'endTime'        : 1355749690.0,
            'condDict'       : { 'Directory' : [ '/lhcb/data', '/lhcb/LHCb' ] }
          }

        returns S_OK / S_ERROR
          { 'graphDataDict' : { '/lhcb/data' : { 1355616000L : 4.9353885242469104,
                                                 1355702400L : 4.8438444870748203 },
                                '/lhcb/LHCb' : { 1355616000L : 3.93538852424691,
                                                 1355702400L : 3.8438444870748198 }
                               },
            'data'          : { '/lhcb/data' : { 1355616000L : 4935388.5242469106,
                                                 1355702400L : 4843844.4870748203 },
                                '/lhcb/LHCb' : { 1355616000L : 3935388.5242469101,
                                                 1355702400L : 3843844.4870748199 }
                               },
            'unit'          : 'TB',
            'granularity'   : 86400
           }
        """

        if reportRequest["grouping"] == "StorageElement":
            return S_ERROR("Grouping by storage element when requesting lfn info makes no sense")

        selectField = self._getSelectStringForGrouping(reportRequest["groupingFields"])
        selectFields = (
            selectField + ", %s, %s, SUM(%s)/SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "LogicalSize", "entriesInBucket"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {"convertToGranularity": "average", "checkNone": True},
        )
        if not retVal["OK"]:
            return retVal

        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)

        accumMaxVal = self._getAccumulationMaxValue(dataDict)
        suitableUnits = self._findSuitableUnit(dataDict, accumMaxVal, "bytes")

        # 3rd value, maxValue is not used
        baseDataDict, graphDataDict, __, unitName = suitableUnits

        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotCatalogSpace(self, reportRequest, plotInfo, filename):
        """Creates <filename>.png file containing information regarding the LFN
        size and the catalog space.

        :param reportRequest: <dict>
           { 'grouping'       : 'Directory',
             'groupingFields' : ( '%s', [ 'Directory' ] ),
             'startTime'      : 1355663249.0,
             'endTime'        : 1355749690.0,
             'condDict'       : { 'Directory' : [ '/lhcb/data', '/lhcb/LHCb' ] }
           }
        :param plotInfo: <dict> ( output of _reportCatalogSpace )
           { 'graphDataDict' : { '/lhcb/data' : { 1355616000L : 4.9353885242469104,
                                                  1355702400L : 4.8438444870748203 },
                                 '/lhcb/LHCb' : { 1355616000L : 3.93538852424691,
                                                  1355702400L : 3.8438444870748198 }
                                },
             'data'          : { '/lhcb/data' : { 1355616000L : 4935388.5242469106,
                                                  1355702400L : 4843844.4870748203 },
                                 '/lhcb/LHCb' : { 1355616000L : 3935388.5242469101,
                                                  1355702400L : 3843844.4870748199 }
                                },
             'unit'          : 'TB',
             'granularity'   : 86400
            }
        :param filename: <str>
          '_plotCatalogSpace'

        returns S_OK / S_ERROR
           { 'plot': True, 'thumbnail': False }
        """

        startEpoch = reportRequest["startTime"]
        endEpoch = reportRequest["endTime"]
        granularity = plotInfo["granularity"]
        dataDict = plotInfo["graphDataDict"]

        metadata = {
            "title": f"LFN space usage by {reportRequest['grouping']}",
            "starttime": startEpoch,
            "endtime": endEpoch,
            "span": granularity,
            "ylabel": plotInfo["unit"],
        }

        dataDict = self._fillWithZero(granularity, startEpoch, endEpoch, dataDict)
        return self._generateStackedLinePlot(filename, dataDict, metadata)

    # .............................................................................
    # catalog Files

    _reportCatalogFilesName = "LFN files"

    def _reportCatalogFiles(self, reportRequest):
        """Reports about the LFN files and the catalog files from the accounting.

        :param reportRequest: <dict>
          { 'grouping'       : 'Directory',
            'groupingFields' : ( '%s', [ 'Directory' ] ),
            'startTime'      : 1355663249.0,
            'endTime'        : 1355749690.0,
            'condDict'       : { 'Directory' : [ '/lhcb/data', '/lhcb/LHCb' ] }
           }

        returns S_OK / S_ERROR
          { 'graphDataDict' : { '/lhcb/data' : { 1355616000L : 4935388.5242469106,
                                                 1355702400L : 4843844.4870748203 },
                                '/lhcb/LHCb' : { 1355616000L : 3935388.5242469101,
                                                 1355702400L : 3843844.4870748199 }
                               },
            'data'          : { '/lhcb/data' : { 1355616000L : 4935388524246.9102,
                                                 1355702400L : 4843844487074.8203 },
                                '/lhcb/LHCb' : { 1355616000L : 3935388524246.9102,
                                                 1355702400L : 3843844487074.8198 }
                               },
            'unit'          : 'Mfiles',
            'granularity'   : 86400
           }
        """

        if reportRequest["grouping"] == "StorageElement":
            return S_ERROR("Grouping by storage element when requesting lfn info makes no sense")

        selectField = self._getSelectStringForGrouping(reportRequest["groupingFields"])
        selectFields = (
            selectField + ", %s, %s, SUM(%s)/SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "LogicalFiles", "entriesInBucket"],
        )

        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {"convertToGranularity": "average", "checkNone": True},
        )
        if not retVal["OK"]:
            return retVal

        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)

        accumMaxVal = self._getAccumulationMaxValue(dataDict)
        suitableUnits = self._findSuitableUnit(dataDict, accumMaxVal, "files")

        # 3rd value, maxValue is not used
        baseDataDict, graphDataDict, __, unitName = suitableUnits

        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotCatalogFiles(self, reportRequest, plotInfo, filename):
        """Creates <filename>.png file containing information regarding the LFN
        files and the catalog files.

        :param reportRequest: <dict>
          { 'grouping'       : 'Directory',
            'groupingFields' : ( '%s', [ 'Directory' ] ),
            'startTime'      : 1355663249.0,
            'endTime'        : 1355749690.0,
            'condDict'       : { 'Directory' : [ '/lhcb/data', '/lhcb/LHCb' ] }
           }
        :param plotInfo: <dict> ( output of _reportCatalogFiles )
          { 'graphDataDict' : { '/lhcb/data' : { 1355616000L : 4935388.5242469106,
                                                 1355702400L : 4843844.4870748203 },
                                '/lhcb/LHCb' : { 1355616000L : 3935388.5242469101,
                                                 1355702400L : 3843844.4870748199 }
                              },
            'data'          : { '/lhcb/data' : { 1355616000L : 4935388524246.9102,
                                                 1355702400L : 4843844487074.8203 },
                                '/lhcb/LHCb' : { 1355616000L : 3935388524246.9102,
                                                1355702400L : 3843844487074.8198 }
                                        },
            'unit'          : 'Mfiles',
            'granularity'   : 86400
           }
        :param filename: <str>
          '_plotCatalogFiles'

        returns S_OK / S_ERROR
           { 'plot': True, 'thumbnail': False }
        """
        startEpoch = reportRequest["startTime"]
        endEpoch = reportRequest["endTime"]
        granularity = plotInfo["granularity"]
        dataDict = plotInfo["graphDataDict"]

        metadata = {
            "title": f"Number of LFNs by {reportRequest['grouping']}",
            "starttime": startEpoch,
            "endtime": endEpoch,
            "span": granularity,
            "ylabel": plotInfo["unit"],
        }

        dataDict = self._fillWithZero(granularity, startEpoch, endEpoch, dataDict)
        return self._generateStackedLinePlot(filename, dataDict, metadata)

    # .............................................................................
    # Physical Space

    _reportPhysicalSpaceName = "PFN size"

    def _reportPhysicalSpace(self, reportRequest):
        """Reports about the PFN size and the physical space from the accounting.

        :param reportRequest: <dict>
          { 'grouping'       : 'StorageElement',
            'groupingFields' : ( '%s', [ 'StorageElement' ] ),
            'startTime'      : 1355663249.0,
            'endTime'        : 1355749690.0,
            'condDict'       : { 'StorageElement' : [ 'CERN-ARCHIVE', 'CERN-DST' ] }
           }

        returns S_OK / S_ERROR
          { 'graphDataDict' : { 'CERN-ARCHIVE' : { 1355616000L : 2.34455676781291,
                                                   1355702400L : 2.5445567678129102 },
                                'CERN-DST'     : { 1355616000L : 0.34455676781290995,
                                                   1355702400L : 0.54455676781290996 }
                               },
            'data'          : { 'CERN-ARCHIVE' : { 1355616000L : 2344556.76781291,
                                                   1355702400L : 2544556.76781291 },
                                'CERN-DST'     : { 1355616000L : 344556.76781290997,
                                                   1355702400L : 544556.76781291002 }
                               },
            'unit'          : 'TB',
            'granularity'   : 86400
           }
        """

        selectField = self._getSelectStringForGrouping(reportRequest["groupingFields"])
        selectFields = (
            selectField + ", %s, %s, SUM(%s/%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "PhysicalSize", "entriesInBucket"],
        )

        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {"convertToGranularity": "average", "checkNone": True},
        )
        if not retVal["OK"]:
            return retVal

        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)

        accumMaxVal = self._getAccumulationMaxValue(dataDict)
        suitableUnits = self._findSuitableUnit(dataDict, accumMaxVal, "bytes")

        # 3rd value, maxValue is not used
        baseDataDict, graphDataDict, __, unitName = suitableUnits

        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotPhysicalSpace(self, reportRequest, plotInfo, filename):
        """Creates <filename>.png file containing information regarding the PFN
        size and the physical space.

        :param reportRequest: <dict>
          { 'grouping'       : 'StorageElement',
            'groupingFields' : ( '%s', [ 'StorageElement' ] ),
            'startTime'      : 1355663249.0,
            'endTime'        : 1355749690.0,
            'condDict'       : { 'StorageElement' : [ 'CERN-ARCHIVE', 'CERN-DST' ] }
          }
        :param plotInfo: <dict> ( output of _reportPhysicalSpace )
          { 'graphDataDict' : { 'CERN-ARCHIVE' : { 1355616000L : 2.34455676781291,
                                                   1355702400L : 2.5445567678129102 },
                                'CERN-DST'     : { 1355616000L : 0.34455676781290995,
                                                   1355702400L : 0.54455676781290996 }
                              },
            'data'          : { 'CERN-ARCHIVE' : { 1355616000L : 2344556.76781291,
                                                   1355702400L : 2544556.76781291 },
                                'CERN-DST'     : { 1355616000L : 344556.76781290997,
                                                   1355702400L : 544556.76781291002 }
                              },
            'unit'          : 'TB',
            'granularity'   : 86400
           }
        :param filename: <str>
          '_plotPhysicalSpace'

        returns S_OK / S_ERROR
           { 'plot': True, 'thumbnail': False }
        """

        startEpoch = reportRequest["startTime"]
        endEpoch = reportRequest["endTime"]
        granularity = plotInfo["granularity"]
        dataDict = plotInfo["graphDataDict"]

        metadata = {
            "title": f"PFN space usage by {reportRequest['grouping']}",
            "starttime": startEpoch,
            "endtime": endEpoch,
            "span": granularity,
            "ylabel": plotInfo["unit"],
        }
        dataDict = self._fillWithZero(granularity, startEpoch, endEpoch, dataDict)
        return self._generateStackedLinePlot(filename, dataDict, metadata)

    # .............................................................................
    # physical Files

    _reportPhysicalFilesName = "PFN files"

    def _reportPhysicalFiles(self, reportRequest):
        """Reports about the PFN files and the physical files from the accounting.

        :param reportRequest: <dict>
           { 'grouping'       : 'StorageElement',
             'groupingFields' : ( '%s', [ 'StorageElement' ] ),
             'startTime'      : 1355663249.0,
             'endTime'        : 1355749690.0,
             'condDict'       : { 'StorageElement' : [ 'CERN-ARCHIVE', 'CERN-DST', 'CERN-BUFFER' ] }
            }

        returns S_OK / S_ERROR
          { 'graphDataDict' : { 'CERN-BUFFER'  : { 1355616000L : 250.65890999999999,
                                                   1355702400L : 261.65890999999999 },
                                'CERN-ARCHIVE' : { 1355616000L : 412.65890999999999,
                                                   1355702400L : 413.65890999999999 },
                                'CERN-DST'     : { 1355616000L : 186.65890999999999,
                                                   1355702400L : 187.65890999999999 }
                               },
            'data'          : { 'CERN-BUFFER'  : { 1355616000L : 250658.91,
                                                   1355702400L : 261658.91 },
                                'CERN-ARCHIVE' : { 1355616000L : 412658.90999999997,
                                                   1355702400L : 413658.90999999997 },
                                'CERN-DST'     : { 1355616000L : 186658.91,
                                                   1355702400L : 187658.91 }
                               },
            'unit'          : 'kfiles',
            'granularity'   : 86400
           }
        """

        selectField = self._getSelectStringForGrouping(reportRequest["groupingFields"])
        selectFields = (
            selectField + ", %s, %s, SUM(%s/%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "PhysicalFiles", "entriesInBucket"],
        )

        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {"convertToGranularity": "average", "checkNone": True},
        )
        if not retVal["OK"]:
            return retVal

        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)

        accumMaxVal = self._getAccumulationMaxValue(dataDict)
        suitableUnits = self._findSuitableUnit(dataDict, accumMaxVal, "files")

        # 3rd value, maxValue is not used
        baseDataDict, graphDataDict, __, unitName = suitableUnits

        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotPhysicalFiles(self, reportRequest, plotInfo, filename):
        """Creates <filename>.png file containing information regarding the PFN
        files and the physical files.

        :param reportRequest: <dict>
           { 'grouping'       : 'StorageElement',
             'groupingFields' : ( '%s', [ 'StorageElement' ] ),
             'startTime'      : 1355663249.0,
             'endTime'        : 1355749690.0,
             'condDict'       : { 'StorageElement' : [ 'CERN-ARCHIVE', 'CERN-DST', 'CERN-BUFFER' ] }
            }
        :param plotInfo: <dict> ( output of _reportPhysicalFiles )
          { 'graphDataDict' : { 'CERN-BUFFER'  : { 1355616000L : 250.65890999999999,
                                                   1355702400L : 261.65890999999999 },
                                'CERN-ARCHIVE' : { 1355616000L : 412.65890999999999,
                                                   1355702400L : 413.65890999999999 },
                                'CERN-DST'     : { 1355616000L : 186.65890999999999,
                                                   1355702400L : 187.65890999999999 }
                              },
            'data'          : { 'CERN-BUFFER'  : { 1355616000L : 250658.91,
                                                   1355702400L : 261658.91 },
                                'CERN-ARCHIVE' : { 1355616000L : 412658.90999999997,
                                                   1355702400L : 413658.90999999997 },
                                'CERN-DST'     : { 1355616000L : 186658.91,
                                                   1355702400L : 187658.91 }
                              },
            'unit'          : 'kfiles',
            'granularity'   : 86400
           }
        :param filename: <str>
          '_plotPhysicalFiles'

        return S_OK / S_ERROR
           { 'plot': True, 'thumbnail': False }
        """

        startEpoch = reportRequest["startTime"]
        endEpoch = reportRequest["endTime"]
        granularity = plotInfo["granularity"]
        dataDict = plotInfo["graphDataDict"]

        metadata = {
            "title": f"Number of PFNs by {reportRequest['grouping']}",
            "starttime": startEpoch,
            "endtime": endEpoch,
            "span": granularity,
            "ylabel": plotInfo["unit"],
        }

        dataDict = self._fillWithZero(granularity, startEpoch, endEpoch, dataDict)
        return self._generateStackedLinePlot(filename, dataDict, metadata)

    # .............................................................................
    # PFN vs LFN File Multiplicity

    _reportPFNvsLFNFileMultiplicityName = "PFN/LFN file ratio"

    def _reportPFNvsLFNFileMultiplicity(self, reportRequest):
        logicalField = "LogicalFiles"
        physicalField = "PhysicalFiles"

        return self._multiplicityReport(reportRequest, logicalField, physicalField)

    def _plotPFNvsLFNFileMultiplicity(self, reportRequest, plotInfo, filename):
        startEpoch = reportRequest["startTime"]
        endEpoch = reportRequest["endTime"]
        granularity = plotInfo["granularity"]
        dataDict = plotInfo["graphDataDict"]

        metadata = {
            "title": f"Ratio of PFN/LFN files by {reportRequest['grouping']}",
            "starttime": startEpoch,
            "endtime": endEpoch,
            "span": granularity,
            "ylabel": plotInfo["unit"],
        }

        dataDict = self._fillWithZero(granularity, startEpoch, endEpoch, dataDict)
        return self._generateStackedLinePlot(filename, dataDict, metadata)

    # .............................................................................
    # PFN vs LFN Size Multiplicity

    _reportPFNvsLFNSizeMultiplicityName = "PFN/LFN size ratio"

    def _reportPFNvsLFNSizeMultiplicity(self, reportRequest):
        logicalField = "LogicalSize"
        physicalField = "PhysicalSize"

        return self._multiplicityReport(reportRequest, logicalField, physicalField)

    def _plotPFNvsLFNSizeMultiplicity(self, reportRequest, plotInfo, filename):
        startEpoch = reportRequest["startTime"]
        endEpoch = reportRequest["endTime"]
        granularity = plotInfo["granularity"]
        dataDict = plotInfo["graphDataDict"]

        metadata = {
            "title": f"Ratio of PFN/LFN space used by {reportRequest['grouping']}",
            "starttime": startEpoch,
            "endtime": endEpoch,
            "span": granularity,
            "ylabel": plotInfo["unit"],
        }

        dataDict = self._fillWithZero(granularity, startEpoch, endEpoch, dataDict)
        return self._generateStackedLinePlot(filename, dataDict, metadata)

    # .............................................................................
    # helper methods

    def _multiplicityReport(self, reportRequest, logicalField, physicalField):
        # Step 1 get the total LFNs for each bucket
        selectFields = (
            "%s, %s, %s, SUM(%s)/SUM(%s)",
            ["Directory", "startTime", "bucketLength", logicalField, "entriesInBucket"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            ("%s", ["Directory"]),
            {"convertToGranularity": "average", "checkNone": True},
        )
        if not retVal["OK"]:
            return retVal

        # 2nd element ( granularity ) is unused
        dataDict, __ = retVal["Value"]
        self.stripDataField(dataDict, 0)
        bucketTotals = self._getBucketTotals(dataDict)

        # Step 2 get the total PFNs
        _selectField = self._getSelectStringForGrouping(reportRequest["groupingFields"])
        selectFields = (
            _selectField + ", %s, %s, SUM(%s/%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", physicalField, "entriesInBucket"],
        )

        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {"convertToGranularity": "average", "checkNone": True},
        )
        if not retVal["OK"]:
            return retVal

        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)

        # Step 3 divide the PFNs by the total amount of LFNs
        finalData = {}

        # FIXME: TO BE replaced by a faster implementation ( see below )
        for k in dataDict:
            for bt in dataDict[k]:
                if bt in bucketTotals:
                    if k not in finalData:
                        finalData[k] = {}
                    finalData[k][bt] = dataDict[k][bt] / bucketTotals[bt]

        #    for key, bucketTotal in dataDict.items():
        #      for bt in bucketTotal.values():
        #        if bt in bucketTotals:
        #          if key not in finalData:
        #            finalData[ key ] = {}
        #          finalData[ key ][ bt ] = bucketTotal[ bt ] / bucketTotals[ bt ]

        return S_OK({"data": finalData, "graphDataDict": finalData, "granularity": granularity, "unit": "PFN / LFN"})



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Bookkeeping file system."""
import sys
import cmd
import pydoc
import shlex
import argparse

from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers
from LHCbDIRAC.Interfaces.API.DiracLHCb import DiracLHCb
from LHCbDIRAC.BookkeepingSystem.Client.LHCB_BKKDBClient import LHCB_BKKDBClient


class LHCbBookkeepingCLI(cmd.Cmd):
    """class."""

    def __init__(self):
        """constructor."""
        cmd.Cmd.__init__(self)
        self.prompt = "$[/]$"
        self.bk = LHCB_BKKDBClient()
        self.diracAPI = DiracLHCb()
        self.currentPath = "/"
        self.do_setDataQualityFlags("OK")
        self.saveParser = argparse.ArgumentParser(description="Save LFNS", prog="save")
        self.saveParser.add_argument("filename", type=str, help="file name")
        self.saveParser.add_argument("-f", "--format", help="txt or py")
        self.saveParser.add_argument("-n", "--num", help="number of files to be saved")
        self.saveParser.add_argument("-c", "--with-fileCatalog", help="save POOL XML catalog in a given site")
        self.sites = {}
        self.sites = DMSHelpers().getShortSiteNames(withStorage=False, tier=(0, 1))

    #############################################################################
    def addCurrentPath(self, path):
        """add a path."""
        if path[0] != "/" and len(self.currentPath) == 1:
            self.currentPath += path
        else:
            self.currentPath += "/" + path

    #############################################################################
    def __printPrompt(self, path=None):
        """prints the prompt."""
        if path is None:
            path = "/"
        self.prompt = "$[" + path + "]$"

    #############################################################################
    def __bklist(self, path):
        """list a path."""
        return self.bk.list(path)

    def __bkListAll(self, path):
        """list the directory with metadata.

        This is equivalent to ls -a
        """
        if path == "":
            path = self.currentPath
        res = self.bk.list(path)
        for i in res:
            print(i["name"])
            for j in i:
                if j not in ["fullpath", "selection", "expandable", "method", "level", "name"]:
                    print("   ", j, i[j])

    #############################################################################
    def __checkDirectory(self, path):
        """is empty directory."""
        return True if self.bk.list(path) else False

    #############################################################################
    def __rootDirectory(self):
        """root."""
        self.currentPath = "/"
        self.__printPrompt(self.currentPath)

    #############################################################################
    def __oneLevelback(self, logical=False):
        """cd .."""
        path = self.currentPath.split("/")
        if path[0] == "" and path[1] == "":
            if not logical:
                self.currentPath = "/"
                self.__printPrompt(self.currentPath)
            else:
                return "/"
        else:
            newpath = ""
            for i in range(len(path) - 1):
                if path[i] != "":
                    newpath += "/" + path[i]
            if newpath == "":
                newpath = "/"
            if not logical:
                self.currentPath = newpath
                self.__printPrompt(self.currentPath)
            else:
                return newpath

    #############################################################################
    @staticmethod
    def help_ls():
        """provides help."""
        print("Usage: ls [OPTION]... [FILE]...")
        print("List information about the FILEs (the current directory by default).")
        print(" Available options: -a")
        print("Usage: ls or ls -a")
        print("ls -a [FILE]")
        print("Note: You can do paging using the | more")
        print("For example: ls | more")

    @staticmethod
    def help_cd():
        """help cd command."""
        print(" cd <dir>")
        print("cd ..")
        print("cd /")
        print("cd")

    #############################################################################
    def do_ls(self, path):
        """ls command."""
        paging = False
        if path.find("|") > -1:
            tmpPath = path.split("|")
            path = ""
            for i in range(len(tmpPath) - 1):
                path += tmpPath[i].strip()
            paging = True

        text = ""
        if len(path) > 0 and path[0] == "-":
            try:
                if path[1] != "a":
                    print(f"ls: invalid option -- {path[1]}")
                    print("Try `help ls' for more information.")
                else:
                    path = path[2:]
                    self.__bkListAll(path)
            except IndexError as e:
                print("Invalid oprion:", e)
        else:
            if path == "":
                path = self.currentPath
            res = self.__bklist(path)
            for i in sorted(res, key=lambda x: x["fullpath"]):
                if paging:
                    text += i["name"] + "\n"
                else:
                    print(i["name"])
        if paging:
            pydoc.ttypager(text)

    #############################################################################
    def do_list(self, path):
        """list commamd."""
        pass

    #############################################################################
    def do_save(self, command):
        """save command."""
        try:
            args = self.saveParser.parse_args(shlex.split(command))
        except argparse.ArgumentError as exc:
            print(exc)
        except SystemExit:
            self.saveParser.print_help()
            return
        if args.format is None:
            args.format = "txt"

        retVal = self.bk.list(self.currentPath)
        files = None
        if retVal:
            if "FileName" not in retVal[0]:
                print(f"No file found belong to {self.currentPath} bookkeeping path!")
                return
            else:
                files = retVal
        else:
            print(retVal)
            return
        if args.num is not None:
            try:
                nb = int(args.num)
            except ValueError as v:
                print(v)
                return
            nbOfFiles = nb
        else:
            nbOfFiles = len(files)

        lfns = {}
        for i in range(nbOfFiles):
            lfns[files[i]["FileName"]] = files[i]

        if args.with_fileCatalog:
            site = args.with_fileCatalog
            lfnList = list(lfns)
            totalFiles = len(lfnList)
            ff = args.filename.split(".")
            catalog = ff[0] + ".xml"
            retVal = self.diracAPI.getInputDataCatalog(lfnList, site, catalog, True)
            nbofsuccsessful = 0
            if retVal["OK"]:
                slist = {}
                faild = {}
                if "Successful" in retVal["Value"]:
                    slist = retVal["Value"]["Successful"]
                if "Failed" in retVal["Value"]:
                    faild = retVal["Value"]["Failed"]
                nbofsuccsessful = len(slist)
                nboffaild = len(faild)
                exist = {}
                for i in slist:
                    exist[i] = lfns[i]

                self.bk.writeJobOptions(exist, args.filename, savedType=None, catalog=catalog, savePfn=slist)

                message = "Total files:" + str(totalFiles) + "\n"
                if site is not None:
                    if nbofsuccsessful:
                        message += str(nbofsuccsessful) + " found " + site + "\n"
                    if nboffaild:
                        message += str(nboffaild) + " not found " + site
                print(message)
                return
        if args.format == "txt":
            text = self.bk.writeJobOptions(lfns, "", args.format)
        elif args.format == "py":
            text = self.bk.writeJobOptions(lfns, args.filename, args.format)
        else:
            raise ValueError(f"Invalid format: {args.format}")

        with open(args.filename, "w") as f:
            f.write(text)

    #############################################################################
    def do_cd(self, path):
        """cd command."""
        newpath = self.currentPath + "/" + path
        if path == "":
            self.currentPath = "/"
            self.__printPrompt(self.currentPath)
        elif path == "..":
            self.__oneLevelback()
        elif path == "/":
            self.__rootDirectory()
        elif path[0] == "/":
            if self.__checkDirectory(path):
                self.currentPath = path
                self.__printPrompt(self.currentPath)
            else:
                print("No such file or directory.")
        elif self.__checkDirectory(newpath):
            self.addCurrentPath(path)
            self.__printPrompt(self.currentPath)
        else:
            print("No such file or directory.")

    #############################################################################
    def do_pwd(self, path):
        """pwd command."""
        print(self.currentPath)

    #############################################################################
    def do_queries(self, command=""):
        """execute query."""
        retVal = self.bk.getPossibleParameters()
        print("The following bookkeeping query types are available:")
        for i in retVal:
            print(" ".rjust(10) + i)
        print("You can change the query types using the 'use' command")

    #############################################################################
    def do_use(self, command):
        """use command."""
        self.bk.setParameter(str(command))
        self.do_cd("/")

    #############################################################################
    @staticmethod
    def help_use():
        """help of use command."""
        print("Usage:")
        print("  use type".rjust(10))
        print("Arguments:")
        print(" type: bookkeeping query type".rjust(10))
        print("The 'type' can be found using the 'queries' command!")
        print("EXAMPE:")
        print(" ".rjust(10) + "use 'Event type'")

    #############################################################################
    @staticmethod
    def help_queries():
        """help of queries command."""
        print("This method shows the available query types!")
        print("Usage:")
        print("  queries")
        print("You can choose a query type using the 'use' command  ")
        print(" NOTE: the default query type is 'Configuration'")

    #############################################################################
    def do_advanceQuery(self, command=""):
        """advancedQuery command."""
        self.bk.setAdvancedQueries(True)
        self.do_cd("/")

    #############################################################################
    @staticmethod
    def help_advanceQuery():
        """help."""
        print("It allows to see more level of the Bookkeeping Tree")
        print("Usage:")
        print("   advanceQuery")

    #############################################################################
    def do_standardQuery(self, command=""):
        """command."""
        self.bk.setAdvancedQueries(False)
        self.do_cd("/")

    #############################################################################
    @staticmethod
    def help_standardQuery():
        """help."""
        print("This is used by default")
        print("It shows a reduced bookkeeping path.")
        print("Usage:")
        print("   standardQuery")

    #############################################################################
    def do_dataQuality(self, command=""):
        """command."""
        print("The following Data Quality flags are available in the bookkeeping!")
        retVal = self.bk.getAvailableDataQuality()
        if retVal["OK"]:
            for i in retVal["Value"]:
                print(" ".ljust(10) + i)
        else:
            print(retVal["Message"])
        print("To set the data quality flags you have to use 'setDataQualityFlags' command!")
        print("More information: 'help setDataQualityFlags'")

    #############################################################################
    @staticmethod
    def help_dataQuality():
        """help."""
        print("This command shows the available data quality flags.")
        print("Usage:")
        print("  dataQuality")
        print("To change the data quality flag use the setDataQualityFlags command")

    #############################################################################
    def do_setDataQualityFlags(self, command):
        """command."""
        qualities = command.split(" ")
        if len(qualities) > 0:
            dataquality = {}
            for i in qualities:
                dataquality[i] = True
            self.bk.setDataQualities(dataquality)
        else:
            print("ERROR: Please give a data quality flag!")

    #############################################################################
    def __moreInfoProcpass(self, command):
        """more information of a directory."""
        found = False
        retVal = self.bk.getProcessingPassSteps({"StepName": command})
        if retVal["OK"]:
            proc = retVal["Value"]
            print(f"{proc['TotalRecords']} {command} step found in the bkk")
            for i in proc["Records"]:
                print(" ".ljust(5) + i)
                for j in proc["Records"][i]:
                    print(" ".ljust(10) + str(j[0]) + ":" + str(j[1]))
                found = True
        else:
            print("ERROR: ", retVal["Message"])
        return found

    #############################################################################
    def do_moreinfo(self, command=""):
        """more info command."""
        if command == "":
            previouspath = self.__oneLevelback(self.currentPath)
            values = self.__bklist(previouspath)
            name = self.currentPath.split("/")
            for i in values:
                if "level" in i and i["level"] == "FileTypes":
                    path = self.currentPath
                    retVal = self.bk.getLimitedFiles({"fullpath": str(path)}, ["nb"], -1, -1)
                    if not retVal["OK"]:
                        print(retVal["Message"])
                        return
                    print("The selected dataset is:")
                    for selected in retVal["Value"]["Extras"]["Selection"]:
                        print("".ljust(5) + selected + " " + str(retVal["Value"]["Extras"]["Selection"][selected]))
                    print("Statistics:")
                    print(" ".ljust(5) + "Number of files:" + str(retVal["Value"]["TotalRecords"]))
                    for selected in retVal["Value"]["Extras"]["GlobalStatistics"]:
                        print(
                            "".ljust(5) + selected + " " + str(retVal["Value"]["Extras"]["GlobalStatistics"][selected])
                        )
                    break
                if i["name"] == name[len(name) - 1]:
                    for j in i:
                        if j not in ["fullpath", "selection", "expandable", "method", "level", "name"]:
                            print("   ", j, i[j])
                if "level" in i and i["level"] == "Processing Pass":
                    found = self.__moreInfoProcpass(command)
                    break
        else:
            values = self.__bklist(self.currentPath)
            found = False
            for i in values:
                if i["name"] == command:
                    for j in i:
                        if j not in ["fullpath", "selection", "expandable", "method", "level", "name"]:
                            print("   ", j, i[j])
                    found = True
                if "level" in i and i["level"] == "Processing Pass":
                    found = self.__moreInfoProcpass(command)
                    break

            if not found:
                print(f" The '{command}' does not found")

    def do_sites(self, command):
        print(f"T0/1 sites: {','.join(site for site in self.sites)}")

    @staticmethod
    def help_sites(self):
        """help."""
        print("it returns a list of T1 sites")

    #############################################################################
    @staticmethod
    def help_setDataQualityFlags():
        """help."""
        print("This command allows to use different data quality flags.")
        print("Usage:")
        print("  setDataQualityFlags flag1 [flag2 flag3, ... flagN]")
        print("Arguments:")
        print("  flag[1...N]:  Data quality flags.")
        print("For example:")
        print(" ".ljust(10) + "setDataQualityFlags OK UNCHECKED")

    #############################################################################
    @staticmethod
    def help_EOF():
        """quit."""
        print("Quits the program")

    #############################################################################
    def do_EOF(self, line):
        """quit command."""
        sys.exit()

    #############################################################################
    def help_save(self):
        """help."""
        self.saveParser.print_help()

    #############################################################################
    @staticmethod
    def help_moreinfo():
        """help method."""
        print("Display the statistics of the selected data.")
        print("Usage:")
        print("  moreinfo")



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""DMScript is a class that creates default switches for DM scripts, decodes
them and sets flags The module also provides a function for printing pretty
results from DMS queries."""
import os
import sys
import time
import tempfile

import DIRAC
from DIRAC import gLogger, gConfig
from DIRAC.Core.Base.Script import Script
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import resolveSEGroup

from LHCbDIRAC.BookkeepingSystem.Client.BKQuery import BKQuery
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient


def __printDictionary(dictionary, offset=0, shift=0, empty="Empty directory", depth=9999):
    """Dictionary pretty printing."""
    key_max = 0
    value_max = 0
    for key, value in dictionary.items():
        key_max = max(key_max, len(str(key)))
        value_max = max(value_max, len(str(value)))
    center = key_max + offset
    newOffset = offset + (shift if shift else key_max)
    for key in sorted(dictionary):
        value = dictionary[key]
        if isinstance(value, dict):
            if not depth:
                value = list(value)
            elif value != {}:
                gLogger.notice(f"{offset * ' '}{key} : ")
                __printDictionary(value, offset=newOffset, shift=shift, empty=empty, depth=depth - 1)
            elif key not in ("Failed", "Successful"):
                gLogger.notice(f"{offset * ' '}{key} : {empty}")
        if isinstance(value, (list, set)):
            if not value:
                gLogger.notice(f"{offset * ' '}{key} : {'[]'}")
            else:
                gLogger.notice(f"{offset * ' '}{key} : ")
                for val in sorted(value):
                    gLogger.notice(f"{newOffset * ' '}{val}")
        elif not isinstance(value, dict):
            # In case value contains \n, indent the lines
            toPrint = str(value).split("\n")
            gLogger.notice(f"{str(key).rjust(center)} : {toPrint.pop(0)}")
            for line in toPrint:
                gLogger.notice((center + 3) * " " + line)


def printDMResult(result, shift=4, empty="Empty directory", script=None, depth=999, offset=0):
    """Printing results returned with 'Successful' and 'Failed' items."""
    if not script:
        script = Script.scriptName
    try:
        if result["OK"]:
            __printDictionary(result["Value"], offset=offset, shift=shift, empty=empty, depth=depth)
            if result["Value"].get("Failed"):
                return 1
            return 0
        gLogger.notice(f"Error in {script} :", result["Message"])
        return 2
    except Exception as e:  # pylint: disable=broad-except
        gLogger.notice(f"Exception while printing results in {script} - Results:", repr(e))
        gLogger.notice(result)
        return 2


class ProgressBar:
    """This object prints a title and a progress bar on stderr."""

    def __init__(self, items, width=None, title=None, chunk=None, step=None, interactive=None, log=None):
        """

        :param items: Number of items to enumerate
        :type items: integer
        :param width: width in co.lumns of the progress bar
        :type width: int
        :param title: Title of the progress bar
        :type title: string
        :param chunk: incremental value by which the counter is incremented (in number of items)
        :type chunk: int
        :param step: frequency at which the bar is updated
        :type step: int
        :param interactive: if False, no printout
        :type interactive: boolean
        """
        if title is None:
            title = ""
        else:
            title += " "
        if width is None:
            width = 50
        if chunk is None:
            chunk = 1
        if step is None:
            step = 1
        if interactive is None:
            interactive = True
        self._log = log if not interactive else None
        self._step = step
        self._width = width
        self._loopNumber = 0
        self._itemCounter = 0
        self._items = items
        self._chunk = chunk
        self._startTime = time.time()
        self._progress = 0
        self._showBar = bool(items > chunk) and bool(items > step) and interactive and sys.stderr.isatty()
        self._interactive = interactive
        self._title = title
        self._backspace = 0
        self._writeTitle()

    def _writeTitle(self):
        """Write the progress bar title to stderr."""
        if not self._interactive:
            return
        if self._showBar:
            sys.stderr.write(f"{self._title}|{self._width * ' '}|")
            self._backspace = self._width + 1
        else:
            sys.stderr.write(self._title)
        sys.stderr.flush()

    def loop(self, increment=True):
        """Called at each iteration of the loop If the iteration modulo "step" is
        0, update the bar Increment the counter of items by "chunk"."""
        if not self._interactive:
            return
        showBar = self._showBar and (self._loopNumber % self._step) == 0
        fraction = min(float(self._itemCounter) / float(self._items), 1.0)
        if increment:
            self._loopNumber += 1
            self._itemCounter += self._chunk
        else:
            showBar = self._showBar
        if showBar:
            progress = int(round(self._width * fraction))
            elapsed = time.time() - self._startTime
            if elapsed > 30.0 and fraction:
                rest = int(elapsed * (1 - fraction) / fraction)
                estimate = " (%d seconds left)" % rest
            else:
                estimate = ""
            blockBlue = "\033[46m"
            endblock = "\033[0m"
            sys.stderr.write(
                "%s%s%s| %5.1f%%%s\033[K"
                % (
                    self._backspace * "\b",
                    blockBlue + (progress - self._progress) * " " + endblock,
                    (self._width - progress) * " ",
                    100.0 * fraction,
                    estimate,
                )
            )
            self._backspace = self._width + 8 - progress + len(estimate)
            self._progress = progress
            sys.stderr.flush()

    def endLoop(self, message=None, timing=True):
        """Closes the progress bar, printing a message and/or the timing since the
        bar was created."""
        if message is None:
            message = "completed"
        timingMsg = f" in {time.time() - self._startTime:.1f} seconds"
        if not self._interactive:
            # Print out message if defined
            if self._log:
                self._log(f"{self._title} {message} {timingMsg}")
            return
        if self._showBar:
            sys.stderr.write("{}\033[K: {}".format((self._progress + self._backspace + 1) * "\b", message))
        if timing:
            if not self._showBar:
                sys.stderr.write(f": {message}")
            sys.stderr.write(timingMsg)
        sys.stderr.write("\n")
        sys.stderr.flush()

    def comment(self, message, optMsg=""):
        """Print a comment."""
        fullMsg = "\n" + message + f" {optMsg}" if optMsg else ""
        gLogger.notice(fullMsg)
        self._writeTitle()
        self.loop(increment=False)


class DMScript:
    """DMScript is a class that creates default switches for DM scripts, decodes
    them and sets flags."""

    def __init__(self):
        """c'tor."""
        self.bkClient = BookkeepingClient()
        self.exceptFileTypes = []
        self.bkClientQuery = None
        self.bkClientQueryDict = {}
        self.options = {}
        self.lastFile = os.path.join(tempfile.gettempdir(), "%d.lastLFNs" % os.getppid())
        self.setLastFile = False
        self.voName = None

    def __voName(self):
        """Returns the name of the VO."""
        if self.voName is None:
            self.voName = gConfig.getValue("/DIRAC/VirtualOrganization", "")
        gLogger.verbose("VO", self.voName)
        return self.voName

    def registerDMSwitches(self):
        """Register switches related to data management, including BK."""
        self.registerBKSwitches()
        self.registerNamespaceSwitches()
        self.registerSiteSwitches()
        self.registerFileSwitches()

    def registerBKSwitches(self):
        """Register switches related to bookkeeping."""
        # BK query switches
        Script.registerSwitch("B:", "BKQuery=", "   Bookkeeping query path", self.setBKQuery)
        Script.registerSwitch(
            "f:",
            "FileType=",
            "   File type (comma separated list, to be used with --Production) [All]",
            self.setFileType,
        )
        Script.registerSwitch(
            "", "ExceptFileType=", "   Exclude the (list of) file types when all are requested", self.setExceptFileTypes
        )
        Script.registerSwitch("", "EventType=", "   Event type", self.setEventType)
        Script.registerSwitch("r:", "Runs=", "   Run or range of runs (r1:r2)", self.setRuns)
        Script.registerSwitch(
            "P:", "Productions=", "   Production ID to search (comma separated list)", self.setProductions
        )
        Script.registerSwitch("", "DQFlags=", "   DQ flag used in query", self.setDQFlags)
        Script.registerSwitch("", "StartDate=", "   Start date for the BK query", self.setStartDate)
        Script.registerSwitch("", "EndDate=", "   End date for the BK query", self.setEndDate)
        Script.registerSwitch("", "Visibility=", "   Required visibility (Yes, No, All) [Yes]", self.setVisibility)
        Script.registerSwitch("", "ReplicaFlag=", "   Required replica flag (Yes, No, All) [Yes]", self.setReplicaFlag)
        Script.registerSwitch("", "TCK=", "   Get files with a given TCK", self.setTCK)
        Script.registerSwitch(
            "", "SMOG2=", "   Required SMOG2 (comma separated list, may include 'Undefined')", self.setSMOG2
        )
        Script.registerSwitch(
            "", "ExtendedDQOK=", "   Comma separated list of (extended) systems which must be ok", self.setExtendedDQOK
        )

    def registerNamespaceSwitches(self, action="search [ALL]"):
        """Register namespace switches."""
        Script.registerSwitch("D:", "Directory=", "   Directory to " + action, self.setDirectory)

    def registerSiteSwitches(self):
        """SE switches."""
        Script.registerSwitch("g:", "Sites=", "  Sites to consider [ALL] (comma separated list)", self.setSites)
        Script.registerSwitch("S:", "SEs=", "  SEs to consider [ALL] (comma separated list)", self.setSEs)

    def registerFileSwitches(self):
        """File switches."""
        Script.registerSwitch("", "File=", "File containing list of LFNs", self.setLFNsFromFile)
        Script.registerSwitch("l:", "LFNs=", "List of LFNs (comma separated)", self.setLFNs)
        Script.registerSwitch("", "Terminal", "LFNs are entered from stdin (--File /dev/stdin)", self.setLFNsFromTerm)
        Script.registerSwitch("", "LastLFNs", "Use last set of LFNs", self.setLFNsFromLast)

    def registerJobsSwitches(self):
        """Job switches."""
        Script.registerSwitch("", "File=", "File containing list of DIRAC jobIds", self.setJobidsFromFile)
        Script.registerSwitch("j:", "DIRACJobids=", "List of DIRAC Jobids (comma separated)", self.setJobids)
        Script.registerSwitch(
            "", "Terminal", "DIRAC Jobids are entered from stdin (--File /dev/stdin)", self.setJobidsFromTerm
        )

    def setProductions(self, arg):
        """Parse production numbers."""
        prods = []
        if arg.upper() == "ALL":
            self.options["Productions"] = arg
            return DIRAC.S_OK()
        try:
            for prod in arg.split(","):
                if prod.find(":") > 0:
                    pr = prod.split(":")
                    for i in range(int(pr[0]), int(pr[1]) + 1):
                        prods.append(i)
                else:
                    prods.append(prod)
            self.options["Productions"] = [int(prod) for prod in prods]
        except ValueError as e:
            gLogger.error(f"Invalid production switch value: {arg}", repr(e))
            self.options["Productions"] = ["Invalid"]
            return DIRAC.S_ERROR()
        return DIRAC.S_OK()

    def setStartDate(self, arg):
        """Setter."""
        self.options["StartDate"] = arg
        return DIRAC.S_OK()

    def setEndDate(self, arg):
        """Setter."""
        self.options["EndDate"] = arg
        return DIRAC.S_OK()

    def setFileType(self, arg):
        """Setter."""
        fileTypes = arg.split(",")
        self.options["FileType"] = fileTypes
        return DIRAC.S_OK()

    def setEventType(self, arg):
        """Setter."""
        eventTypes = arg.split(",")
        self.options["EventType"] = eventTypes
        return DIRAC.S_OK()

    def setExceptFileTypes(self, arg):
        """Setter."""
        self.exceptFileTypes += arg.split(",")
        return DIRAC.S_OK()

    def setBKQuery(self, arg):
        """Setter."""
        # BKQuery could either be a BK path or a file path that contains the BK items
        self.bkClientQuery = None
        self.bkClientQueryDict = {}
        self.options["BKPath"] = arg
        return DIRAC.S_OK()

    def setRuns(self, arg):
        """Setter."""
        self.options["Runs"] = arg
        return DIRAC.S_OK()

    def setDQFlags(self, arg):
        """Setter."""
        dqFlags = arg.split(",")
        self.options["DQFlags"] = dqFlags
        return DIRAC.S_OK()

    def setTCK(self, arg):
        """Setter."""
        tcks = arg.split(",")
        self.options["TCK"] = tcks
        return DIRAC.S_OK()

    def setSMOG2(self, arg):
        """Setter."""
        states = arg.split(",")
        self.options["SMOG2"] = states
        return DIRAC.S_OK()

    def setExtendedDQOK(self, arg):
        """Setter."""
        states = arg.split(",")
        self.options["ExtendedDQOK"] = states
        return DIRAC.S_OK()

    def setVisibility(self, arg):
        """Setter."""
        if arg.lower() in ("yes", "no", "all"):
            self.options["Visibility"] = arg
        else:
            gLogger.error(f"Unknown visibility flag: {arg}")
            return DIRAC.exit(1)
        return DIRAC.S_OK()

    def setReplicaFlag(self, arg):
        """Setter."""
        if arg.lower() in ("yes", "no", "all"):
            self.options["ReplicaFlag"] = arg
        else:
            gLogger.error(f"Unknown replica flag: {arg}")
            return DIRAC.exit(1)
        return DIRAC.S_OK()

    def setDirectory(self, arg):
        """Setter."""
        if os.path.exists(arg) and not os.path.isdir(arg):
            with open(arg) as inFile:
                directories = [line.split()[0] for line in inFile.read().splitlines() if line.strip()]
        else:
            directories = arg.split(",")
        self.options.setdefault("Directory", set()).update(directories)
        return DIRAC.S_OK()

    def setSites(self, arg):
        """Setter."""
        siteShortNames = {
            "CERN": "LCG.CERN.cern",
            "CNAF": "LCG.CNAF.it",
            "GRIDKA": "LCG.GRIDKA.de",
            "NIKHEF": "LCG.NIKHEF.nl",
            "SARA": "LCG.SARA.nl",
            "PIC": "LCG.PIC.es",
            "RAL": "LCG.RAL.uk",
            "IN2P3": "LCG.IN2P3.fr",
            "NCBJ": "LCG.NCBJ.pl",
        }
        sites = arg.split(",")
        self.options["Sites"] = [siteShortNames.get(site.upper(), site) for site in sites]
        return DIRAC.S_OK()

    def setSEs(self, arg):
        """Setter."""
        self.options["SEs"] = arg.split(",")
        return DIRAC.S_OK()

    def setLFNs(self, arg):
        """Setter."""
        if arg:
            self.options.setdefault("LFNs", set()).update(arg.split(","))
        return DIRAC.S_OK()

    def setLFNsFromTerm(self, arg=None):
        """Setter."""
        return self.setLFNsFromFile(arg)

    def getLFNsFromList(self, lfns, directories=False):
        """Returns a list of LFNs from a list of strings LFNs start with the last
        occurence of /<vo>/ in the file name and ends with a set of delimiters If
        directories is True, only normalized directories (ending with a "/" are
        returned."""
        if isinstance(lfns, str):
            lfnList = lfns.strip().split(",")
        elif isinstance(lfns, (list, set, dict)):
            lfnList = [lfn.strip() for lfn1 in lfns for lfn in lfn1.split(",")]
        else:
            gLogger.error(f"getLFNsFromList: invalid type {type(lfns)}")
            return []
        vo = self.__voName()
        if vo:
            vo = f"/{vo}"
            lfnList = [x.replace("LFN: ", "LFN  ") for x in lfnList]
            lfnList = [x.split("LFN:")[-1].strip() for x in lfnList]
            for sep in ('"', ",", "'", ":", "(", ")", ";", "|"):
                lfnList = [x.replace(sep, " ") for x in lfnList]
            lfnList = [vo + lfn.split(vo)[-1].split()[0] if vo in lfn else lfn if lfn == vo else "" for lfn in lfnList]
            lfnList = [lfn.split("?")[0] for lfn in lfnList]
            lfnList = (
                [lfn for lfn in lfnList if lfn.endswith("/")]
                if directories
                else [lfn for lfn in lfnList if not lfn.endswith("/")]
            )
        return sorted(lfn for lfn in set(lfnList) if lfn)

    @staticmethod
    def getJobIDsFromList(jobids):
        """it returns a list of jobids using a string."""
        jobidsList = []
        if isinstance(jobids, str):
            jobidsList = jobids.split(",")
        elif isinstance(jobids, list):
            jobidsList = [jobid for jobid1 in jobids for jobid in jobid1.split(",")]
        jobidsList = [jobid for jobid in jobidsList if jobid]
        return jobidsList

    def setLFNsFromLast(self, _val):
        """Setter when --Last is used."""
        if os.path.exists(self.lastFile):
            return self.setLFNsFromFile(self.lastFile)
        gLogger.fatal(f"Last file {self.lastFile} does not exist")
        DIRAC.exit(2)

    def setLFNsFromFile(self, arg):
        """Reads the content of a file or from stdin (in which case a temporary
        file will be created) LFNs are not parsed at this stage."""
        if isinstance(arg, str) and arg.lower() == "last":
            arg = self.lastFile
        # Make a list of files
        if isinstance(arg, str):
            files = arg.split(",")
        elif isinstance(arg, list):
            files = arg
        elif not arg:
            files = [arg]
        else:
            raise ValueError(f"Invalid argument type {type(arg)}")
        nfiles = 0
        for fName in files:
            try:
                with open(fName if fName else "/dev/stdin") as inFile:
                    lfns = inFile.read().splitlines()
                nfiles += len(lfns)
            except (EOFError, OSError):
                lfns = fName.split(",")
            self.options.setdefault("LFNs", set()).update(lfns)
        if nfiles:
            self.setLastFile = arg if arg else "term"
        return DIRAC.S_OK()

    def getOptions(self):
        """Returns all options."""
        return self.options

    def getOption(self, switch, default=None):
        """Get a specific items set by the setters."""
        if switch == "SEs":
            # SEs have to be resolved recursively using StorageElementGroups
            return resolveSEGroup(self.options.get(switch, default))
        value = self.options.get(switch, default)
        if switch in ("LFNs", "Directory"):
            # Special case for getting LFNs or directories: parse the option
            if value == default and switch == "Directory":
                value = self.options.get("LFNs", default)
            if not value:
                if not sys.stdin.isatty():
                    # If the input file is a pipe, no need to specify it
                    self.setLFNsFromTerm()
                    value = self.options.get("LFNs", default)
            if value:
                # Parse the LFNs out of the "LFNs" option list
                value = self.getLFNsFromList(value, directories=switch == "Directory")
            if value and self.setLastFile and switch == "LFNs":
                # Storethe list of LFNs in a temporary file
                gLogger.always(f"Got {len(value)} LFNs")
                if self.setLastFile != self.lastFile:
                    self.setLastFile = False
                    with open(self.lastFile, "w") as tmpFile:
                        tmpFile.write("\n".join(sorted(value)))
        if isinstance(value, set):
            # Return a sorted list from a set
            value = sorted(value)
        return value

    def getBKQuery(self, visible=None):
        """Returns a BKQuery object from the requested BK information."""
        mandatoryKeys = {
            ("ConfigName", "ConfigVersion"),
            "Production",
            ("FileType", "RunNumber"),
            ("FileType", "StartRun"),
        }
        if self.bkClientQuery:
            return self.bkClientQuery
        if self.bkClientQueryDict:
            self.bkClientQuery = BKQuery(self.bkClientQueryDict)
        else:
            visible = self.options.get("Visibility", "Yes") if visible is None else visible
            bkPath = self.options.get("BKPath")
            prods = self.options.get("Productions")
            runs = self.options.get("Runs")
            fileTypes = self.options.get("FileType")
            eventTypes = self.options.get("EventType")
            self.bkClientQuery = BKQuery(
                bkQuery=bkPath, prods=prods, runs=runs, fileTypes=fileTypes, eventTypes=eventTypes, visible=visible
            )
        bkQueryDict = self.bkClientQuery.getQueryDict()
        found = False
        for key in mandatoryKeys:
            if isinstance(key, str) and key in bkQueryDict:
                found = True
                break
            elif isinstance(key, (list, tuple)) and not set(key) - set(bkQueryDict):
                found = True
                break
        if not found:
            self.bkClientQuery = None
            return None
        # Add extra requirements
        self.bkClientQuery.setExceptFileTypes(self.exceptFileTypes)
        if "DQFlags" in self.options:
            self.bkClientQuery.setDQFlag(self.options["DQFlags"])
        if "StartDate" in self.options:
            self.bkClientQuery.setOption("StartDate", self.options["StartDate"])
        if "EndDate" in self.options:
            self.bkClientQuery.setOption("EndDate", self.options["EndDate"])
        if "ReplicaFlag" in self.options:
            self.bkClientQuery.setOption("ReplicaFlag", self.options["ReplicaFlag"])
        if "TCK" in self.options:
            self.bkClientQuery.setOption("TCK", self.options["TCK"])
        if "SMOG2" in self.options:
            self.bkClientQuery.setOption("SMOG2", self.options["SMOG2"])
        if "ExtendedDQOK" in self.options:
            self.bkClientQuery.setOption("ExtendedDQOK", self.options["ExtendedDQOK"])
        return self.bkClientQuery

    def getRequestID(self, prod=None):
        """Get the request ID for a single production."""
        from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

        if not prod:
            prod = self.options.get("Productions", [])
        requestID = None
        if isinstance(prod, str):
            prods = [prod]
        else:
            prods = prod
        if len(prods) == 1 and str(prods[0]).upper() != "ALL":
            res = TransformationClient().getTransformation(prods[0])
            if res["OK"]:
                requestID = int(res["Value"]["TransformationFamily"])
        return requestID

    def setJobidsFromFile(self, arg):
        """It fill a list with DIRAC jobids read from a file
        NOTE: The file format is equivalent to the file format when the content is
        a list of LFNs."""
        try:
            with open(arg if arg else "/dev/stdin") as inFile:
                jobids = self.getJobIDsFromList(inFile.read().splitlines())
        except OSError as error:
            gLogger.exception("Reading jobids from a file is failed with exception:", repr(error))
            jobids = self.getJobIDsFromList(arg)
        self.options.setdefault("JobIDs", set()).update(jobids)
        return DIRAC.S_OK()

    def setJobids(self, arg):
        """It fill a list with DIRAC Jobids."""
        if arg:
            self.options.setdefault("JobIDs", set()).update(self.getJobIDsFromList(arg))
        return DIRAC.S_OK()

    def setJobidsFromTerm(self):
        """It is used to fill a list with jobids."""
        return self.setJobidsFromFile(None)



###############################################################################
# (c) Copyright 2022 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
""" TornadoBookkeepingManager is the implementation of the BookkeepingManager service in HTTPS

    .. literalinclude:: ../ConfigTemplate.cfg
      :start-after: ##BEGIN TornadoBookkeepingManager:
      :end-before: ##END
      :dedent: 2
      :caption: TornadoBookkeepingManager options
"""
from DIRAC import S_OK
from DIRAC.Core.Tornado.Server.TornadoService import TornadoService
from LHCbDIRAC.BookkeepingSystem.Client import JEncoder
from LHCbDIRAC.BookkeepingSystem.Service.BookkeepingManagerHandler import BookkeepingManagerHandlerMixin
from LHCbDIRAC.BookkeepingSystem.Service.BookkeepingManagerHandler import default


class TornadoBookkeepingManagerHandler(BookkeepingManagerHandlerMixin, TornadoService):
    def export_streamToClient(self, parameters):
        """This method is used to transfer data using a file.

        Currently two client methods are using this function: getFiles, getFilesWithMetadata
        """

        in_dict = JEncoder.loads(parameters)
        self.log.verbose("Received the following dictionary:", str(in_dict))
        methodName = in_dict["MethodName"]
        if methodName == "getFiles":
            retVal = self._getFiles(in_dict)
        elif methodName == "getFilesWithMetadata":
            retVal = self._getFilesWithMetadata(in_dict)
        else:
            raise NotImplementedError(methodName)

        if not retVal["OK"]:
            self.log.error("Failed to send files:", str(in_dict))
            return retVal

        return S_OK(JEncoder.dumps(retVal["Value"]))



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""BookkeepingManager service is the front-end to the Bookkeeping database."""
from functools import partial

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.Core.Utilities.Decorators import deprecated
from DIRAC.Core.Utilities.ReturnValues import convertToReturnValue, returnValueOrRaise
from DIRAC.ConfigurationSystem.Client.PathFinder import getServiceSection
from DIRAC.ConfigurationSystem.Client.Helpers import cfgPath
from DIRAC.ConfigurationSystem.Client.Config import gConfig


from LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB import OracleBookkeepingDB
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.XMLFilesReaderManager import XMLFilesReaderManager
from LHCbDIRAC.BookkeepingSystem.Client import JEncoder
from LHCbDIRAC.BookkeepingSystem.DB.Utilities import checkEnoughBKArguments

# pylint: disable=invalid-name


default = "ALL"


class BookkeepingManagerHandlerMixin:
    """Bookkeeping Service class.

    It serves the requests made the users by using the BookkeepingClient.
    """

    @classmethod
    def initializeHandler(cls, _serviceInfoDict):
        """Initializes the variables used to identify queries, which are not
        containing enough conditions."""

        cls.bkkDB = OracleBookkeepingDB()
        cls.xmlReader = XMLFilesReaderManager()
        cls.__eventTypeCache = {}

        bkkSection = getServiceSection("Bookkeeping", "BookkeepingManager")
        if not bkkSection:
            cls.email = "lhcb-bookkeeping@cern.ch"
            cls.forceExecution = False
        else:
            cls.email = gConfig.getValue(cfgPath(bkkSection, "Email"), "lhcb-bookkeeping@cern.ch")
            cls.forceExecution = gConfig.getValue(cfgPath(bkkSection, "ForceExecution"), False)
        cls.log.info(f"Email used to track queries: {cls.email} forceExecution", cls.forceExecution)
        return S_OK()

    #############################################################################
    types_sendXMLBookkeepingReport = [str]

    def export_sendXMLBookkeepingReport(self, xml):
        """This method is used to upload an xml report which is produced after when
        the job successfully finished. The input parameter 'xml' is a string which
        contains various information (metadata) about the finished job in the Grid
        in an XML format.

        :param str xml: bookkeeping report
        """
        retVal = self.xmlReader.readXMLfromString(xml)
        if not retVal["OK"]:
            self.log.error("Issue reading XML", retVal["Message"])
        return retVal

    #############################################################################
    types_getAvailableSteps = [dict]

    @classmethod
    def export_getAvailableSteps(cls, in_dict):
        """It returns all the available steps which corresponds to a given
        conditions.

        The in_dict contains the following conditions: StartDate, StepId,
        InputFileTypes, OutputFileTypes,     ApplicationName,
        ApplicationVersion, OptionFiles, DDDB, CONDDB, ExtraPackages,
        Visible, ProcessingPass, Usable, RuntimeProjects, DQTag,
        OptionsFormat, StartItem, MaxItem
        """
        return cls.bkkDB.getAvailableSteps(in_dict)

    #############################################################################
    types_getDataTakingConditionID = [str]

    @classmethod
    def export_getDataTakingConditionID(cls, daq_desc):
        """It returns all the condition ID given the description"""
        return cls.bkkDB._getDataTakingConditionId(daq_desc)

    #############################################################################
    types_getRuntimeProjects = [dict]

    @classmethod
    def export_getRuntimeProjects(cls, in_dict):
        """It returns a runtime project for a given step.

        The input parameter is a in_dictionary which has only one key StepId
        """
        return cls.bkkDB.getRuntimeProjects(in_dict)

    #############################################################################
    types_getStepInputFiles = [int]

    @classmethod
    def export_getStepInputFiles(cls, stepId):
        """It returns the input files for a given step."""
        retVal = cls.bkkDB.getStepInputFiles(stepId)
        if not retVal["OK"]:
            return retVal

        records = [list(record) for record in retVal["Value"]]
        return S_OK({"ParameterNames": ["FileType", "Visible"], "Records": records, "TotalRecords": len(records)})

    #############################################################################
    types_setStepInputFiles = [int, list]

    @classmethod
    def export_setStepInputFiles(cls, stepid, files):
        """It is used to set input file types to a Step."""
        return cls.bkkDB.setStepInputFiles(stepid, files)

    #############################################################################
    types_setStepOutputFiles = [int, list]

    @classmethod
    def export_setStepOutputFiles(cls, stepid, files):
        """It is used to set output file types to a Step."""
        return cls.bkkDB.setStepOutputFiles(stepid, files)

    #############################################################################
    types_getStepOutputFiles = [int]

    @classmethod
    def export_getStepOutputFiles(cls, stepId):
        """It returns the output file types for a given Step."""
        retVal = cls.bkkDB.getStepOutputFiles(stepId)
        if not retVal["OK"]:
            return retVal

        records = []
        parameters = ["FileType", "Visible"]
        for record in retVal["Value"]:
            records += [list(record)]
        return S_OK({"ParameterNames": parameters, "Records": records, "TotalRecords": len(records)})

    #############################################################################
    types_getAvailableFileTypes = []

    @classmethod
    def export_getAvailableFileTypes(cls):
        """more info in the BookkeepingClient.py."""
        return cls.bkkDB.getAvailableFileTypes()

    #############################################################################
    types_insertFileTypes = [str, str, str]

    @classmethod
    def export_insertFileTypes(cls, ftype, desc, fileType):
        """It is used to register a file type. It has the following input
        parameters:

        :param str ftype: file type; for example: COOL.DST
        :param str desc: a short description which describes the file content
        :paran str fileType: the file format such as ROOT, POOL_ROOT, etc.
        """
        return cls.bkkDB.insertFileTypes(ftype, desc, fileType)

    #############################################################################
    types_insertStep = [dict]

    @classmethod
    def export_insertStep(cls, in_dict):
        """It used to insert a step to the Bookkeeping Metadata Catalogue. The
        imput parameter is a dictionary which contains the steps attributes. For
        example: Dictionary format:

        {'Step': {'ApplicationName': 'DaVinci',
        'Usable': 'Yes',
        'StepId': '',
        'ApplicationVersion': 'v29r1', 'ext-comp-1273':
        'CHARM.MDST (Charm micro dst)', 'ExtraPackages': '', 'StepName': 'davinci prb2',
        'ProcessingPass': 'WG-Coool', 'ext-comp-1264': 'CHARM.DST (Charm stream)', 'Visible': 'Y', 'DDDB': '',
        'OptionFiles': '', 'CONDDB': ''}, 'OutputFileTypes': [{'Visible': 'Y', 'FileType': 'CHARM.MDST'}],
        'InputFileTypes': [{'Visible': 'Y', 'FileType': 'CHARM.DST'}],'RuntimeProjects':[{StepId:13878}]}
        """
        return cls.bkkDB.insertStep(in_dict)

    #############################################################################
    types_deleteStep = [int]

    @classmethod
    def export_deleteStep(cls, stepid):
        """It used to delete a given step."""
        return cls.bkkDB.deleteStep(stepid)

    #############################################################################
    types_deleteStepContainer = [int]

    @classmethod
    def export_deleteStepContainer(cls, stepid):
        """It used to delete a given step."""
        return cls.bkkDB.deleteStepContainer(stepid)

    #############################################################################
    types_updateStep = [dict]

    @classmethod
    def export_updateStep(cls, in_dict):
        """It is used to modify the step attributes."""
        return cls.bkkDB.updateStep(in_dict)

    ##############################################################################
    types_getAvailableConfigNames = []

    @classmethod
    def export_getAvailableConfigNames(cls):
        """It returns all the available configuration names which are used."""
        retVal = cls.bkkDB.getAvailableConfigNames()
        if not retVal["OK"]:
            return retVal

        records = [list(record) for record in retVal["Value"]]
        return S_OK({"ParameterNames": ["Configuration Name"], "Records": records, "TotalRecords": len(records)})

    #############################################################################
    types_getConfigVersions = [dict]

    @classmethod
    def export_getConfigVersions(cls, in_dict):
        """It returns all the available configuration version for a given
        condition.

        Input parameter is a dictionary which has the following key: 'ConfigName'
        For example: in_dict = {'ConfigName':'MC'}
        """
        configName = in_dict.get("ConfigName", default)
        retVal = cls.bkkDB.getConfigVersions(configName)
        if not retVal["OK"]:
            return retVal
        records = []
        parameters = ["Configuration Version"]
        for record in retVal["Value"]:
            records += [list(record)]
        return S_OK({"ParameterNames": parameters, "Records": records, "TotalRecords": len(records)})

    #############################################################################
    types_getConditions = [dict]

    @classmethod
    def export_getConditions(cls, in_dict):
        """It returns all the available conditions for a given conditions.

        Input parameter is a dictionary which has the following keys: 'ConfigName', 'ConfigVersion', 'EventType'
        For example: in_dict = {'ConfigName':'MC','ConfigVersion':'MC10'}
        """
        result = S_ERROR()
        ok = True
        configName = in_dict.get("ConfigName", default)
        configVersion = in_dict.get("ConfigVersion", default)
        evt = in_dict.get("EventType", in_dict.get("EventTypeId", default))

        if "EventTypeId" in in_dict:
            cls.log.verbose("EventTypeId will be not accepted! Please change it to EventType")

        retVal = cls.bkkDB.getConditions(configName, configVersion, evt)
        if retVal["OK"]:
            values = retVal["Value"]
            sim_parameters = [
                "SimId",
                "Description",
                "BeamCondition",
                "BeamEnergy",
                "Generator",
                "MagneticField",
                "DetectorCondition",
                "Luminosity",
                "G4settings",
            ]
            daq_parameters = [
                "DaqperiodId",
                "Description",
                "BeamCondition",
                "BeanEnergy",
                "MagneticField",
                "VELO",
                "IT",
                "TT",
                "OT",
                "RICH1",
                "RICH2",
                "SPD_PRS",
                "ECAL",
                "HCAL",
                "MUON",
                "L0",
                "HLT",
                "VeloPosition",
            ]
            sim_records = []
            daq_records = []

            if len(values) > 0:
                for record in values:
                    if record[0] is not None:
                        sim_records += [
                            [
                                record[0],
                                record[2],
                                record[3],
                                record[4],
                                record[5],
                                record[6],
                                record[7],
                                record[8],
                                record[9],
                            ]
                        ]
                    elif record[1] is not None:
                        daq_records += [
                            [
                                record[1],
                                record[10],
                                record[11],
                                record[12],
                                record[13],
                                record[14],
                                record[15],
                                record[16],
                                record[17],
                                record[18],
                                record[19],
                                record[20],
                                record[21],
                                record[22],
                                record[23],
                                record[24],
                                record[25],
                                record[26],
                            ]
                        ]
                    else:
                        result = S_ERROR("Condition does not exist")
                        ok = False
            if ok:
                result = S_OK(
                    [
                        {"ParameterNames": sim_parameters, "Records": sim_records, "TotalRecords": len(sim_records)},
                        {"ParameterNames": daq_parameters, "Records": daq_records, "TotalRecords": len(daq_records)},
                    ]
                )
        else:
            result = retVal

        return result

    #############################################################################
    types_getProcessingPass = [dict, str]

    @classmethod
    def export_getProcessingPass(cls, in_dict, path=None):
        """It returns the processing pass for a given conditions.

        Input parameter is a dictionary and a path (string) which has the following keys:
        'ConfigName', 'ConfigVersion', 'ConditionDescription','Production', 'RunNumber', 'EventType'
        This method is used to recursively browse the processing pass.
        To start the browsing you have to define the path as a root: path = '/'
        Note: it returns a list with two dictionary. First dictionary contains the processing passes
        while the second dictionary contains the event types.
        """
        if path is None:
            path = "/"
        configName = in_dict.get("ConfigName", default)
        configVersion = in_dict.get("ConfigVersion", default)
        conddescription = in_dict.get(
            "ConditionDescription",  # old single
            in_dict.get("SimulationConditions", in_dict.get("DataTakingConditions", default)),
        )
        prod = in_dict.get("Production", default)
        runnb = in_dict.get("RunNumber", default)
        evt = in_dict.get("EventType", in_dict.get("EventTypeId", default))
        return cls.bkkDB.getProcessingPass(configName, configVersion, conddescription, runnb, prod, evt, path)

    #############################################################################
    types_getProductions = [dict]

    @classmethod
    def export_getProductions(cls, in_dict):
        """It returns the productions for a given conditions.

        Input parameter is a dictionary which has the following keys:
        'ConfigName', 'ConfigVersion', 'ConditionDescription',
        'EventType','ProcessingPass'
        """
        configName = in_dict.get("ConfigName", default)
        configVersion = in_dict.get("ConfigVersion", default)
        conddescription = in_dict.get(
            "ConditionDescription",  # old single
            in_dict.get("SimulationConditions", in_dict.get("DataTakingConditions", default)),
        )

        processing = in_dict.get("ProcessingPass", default)
        evt = in_dict.get("EventType", in_dict.get("EventTypeId", default))
        replicaFlag = in_dict.get("ReplicaFlag", "Yes")
        ftype = in_dict.get("FileType", default)
        visible = in_dict.get("Visible", "Y")
        if "EventTypeId" in in_dict:
            cls.log.verbose("The EventTypeId has to be replaced by EventType!")

        retVal = cls.bkkDB.getProductions(
            configName, configVersion, conddescription, processing, evt, visible, ftype, replicaFlag
        )
        if not retVal["OK"]:
            return retVal
        records = []
        parameters = ["Production/RunNumber"]
        for record in retVal["Value"]:
            records += [[record[0]]]
        return S_OK({"ParameterNames": parameters, "Records": records, "TotalRecords": len(records)})

    #############################################################################
    types_getFileTypes = [dict]

    @classmethod
    def export_getFileTypes(cls, in_dict):
        """It returns the file types for a given conditions.

        Input parameter is a dictionary which has the following keys:
        'ConfigName', 'ConfigVersion', 'ConditionDescription', 'EventType','ProcessingPass','Production','RunNumber'
        """
        configName = in_dict.get("ConfigName", default)
        configVersion = in_dict.get("ConfigVersion", default)
        conddescription = in_dict.get(
            "ConditionDescription",  # old single
            in_dict.get("SimulationConditions", in_dict.get("DataTakingConditions", default)),
        )
        processing = in_dict.get("ProcessingPass", default)
        evt = in_dict.get("EventType", in_dict.get("EventTypeId", default))
        production = in_dict.get("Production", default)
        runnb = in_dict.get("RunNumber", default)
        visible = in_dict.get("Visible", "Y")
        replicaflag = in_dict.get("ReplicaFlag", "Yes")

        if "EventTypeId" in in_dict:
            cls.log.verbose("The EventTypeId has to be replaced by EventType!")

        retVal = cls.bkkDB.getFileTypes(
            configName, configVersion, conddescription, processing, evt, runnb, production, visible, replicaflag
        )
        if not retVal["OK"]:
            return retVal
        records = []
        parameters = ["FileTypes"]
        for record in retVal["Value"]:
            records += [[record[0]]]
        return S_OK({"ParameterNames": parameters, "Records": records, "TotalRecords": len(records)})

    #############################################################################
    def transfer_toClient(self, parameters, token, fileHelper):
        """This method is used to transfer data using a file.

        Currently two client methods are using this function: getFiles, getFilesWithMetadata
        """
        in_dict = JEncoder.loads(parameters)
        self.log.verbose("Received the following dictionary:", str(in_dict))
        methodName = in_dict.get("MethodName", default)
        if methodName == "getFiles":
            retVal = self._getFiles(in_dict)
        else:
            retVal = self._getFilesWithMetadata(in_dict)

        if not retVal["OK"] and "ExecInfo" in retVal:
            # ExecInfo isn't encodable by DEncode or JEncode and it's normally dropped
            # transparently by the RPC mechanism. As we're doing it by hand here we need
            # to remove it explicitly.
            retVal.pop("ExecInfo")
        fileString = JEncoder.dumps(retVal)

        retVal = fileHelper.stringToNetwork(fileString)
        if not retVal["OK"]:
            self.log.error("Failed to send files:", f"{in_dict}")
            return retVal
        self.log.debug("Sent files for", f"{in_dict} of size {len(fileString)}")
        return S_OK()

    #############################################################################
    @classmethod
    @checkEnoughBKArguments
    def _getFiles(cls, in_dict):
        """It returns a list of files."""
        simdesc = default
        datataking = in_dict.get(
            "SimulationConditions",  # old primary
            in_dict.get("DataTakingConditions", in_dict.get("ConditionDescription", default)),  # old secondary
        )
        procPass = in_dict.get("ProcessingPass", default)
        ftype = in_dict.get("FileType", default)
        evt = in_dict.get("EventType", default)
        configname = in_dict.get("ConfigName", default)
        configversion = in_dict.get("ConfigVersion", default)
        prod = in_dict.get("Production", in_dict.get("ProductionID", default))
        flag = in_dict.get("DataQuality", in_dict.get("DataQualityFlag", default))
        startd = in_dict.get("StartDate", None)
        endd = in_dict.get("EndDate", None)
        nbofevents = in_dict.get("NbOfEvents", False)
        startRunID = in_dict.get("StartRun", None)
        endRunID = in_dict.get("EndRun", None)
        runNbs = in_dict.get("RunNumber", in_dict.get("RunNumbers", []))
        if not isinstance(runNbs, list):
            runNbs = [runNbs]
        replicaFlag = in_dict.get("ReplicaFlag", "Yes")
        visible = in_dict.get("Visible", default)
        filesize = in_dict.get("FileSize", False)
        tck = in_dict.get("TCK", [])
        jobStart = in_dict.get("JobStartDate", None)
        jobEnd = in_dict.get("JobEndDate", None)
        smog2States = in_dict.get("SMOG2", None)
        extendedDQOK = in_dict.get("ExtendedDQOK", None)

        if "ProductionID" in in_dict:
            cls.log.verbose("ProductionID will be removed. It will changed to Production")

        if "DataQualityFlag" in in_dict:
            cls.log.verbose("DataQualityFlag will be removed. It will changed to DataQuality")

        if "RunNumbers" in in_dict:
            cls.log.verbose("RunNumbers will be removed. It will changed to RunNumber")

        retVal = cls.bkkDB.getFiles(
            simdesc,
            datataking,
            procPass,
            ftype,
            evt,
            configname,
            configversion,
            prod,
            flag,
            startd,
            endd,
            nbofevents,
            startRunID,
            endRunID,
            runNbs,
            replicaFlag,
            visible,
            filesize,
            tck,
            jobStart,
            jobEnd,
            smog2States,
            extendedDQOK,
        )
        if not retVal["OK"]:
            return retVal

        return S_OK([i[0] for i in retVal["Value"]])

    #############################################################################
    @classmethod
    @checkEnoughBKArguments
    def _getFilesWithMetadata(cls, in_dict):
        """It returns the files with their metadata."""
        configName = in_dict.get("ConfigName", default)
        configVersion = in_dict.get("ConfigVersion", default)
        conddescription = in_dict.get(
            "ConditionDescription",  # old single
            in_dict.get("SimulationConditions", in_dict.get("DataTakingConditions", default)),
        )
        processing = in_dict.get("ProcessingPass", default)
        evt = in_dict.get("EventType", in_dict.get("EventTypeId", default))
        production = in_dict.get("Production", default)
        filetype = in_dict.get("FileType", default)
        quality = in_dict.get("DataQuality", in_dict.get("DataQualityFlag", in_dict.get("Quality", default)))
        visible = in_dict.get("Visible", "Y")
        replicaFlag = in_dict.get("ReplicaFlag", "Yes")
        startDate = in_dict.get("StartDate", None)
        endDate = in_dict.get("EndDate", None)
        runnumbers = in_dict.get("RunNumber", in_dict.get("RunNumbers", []))
        startRunID = in_dict.get("StartRun", None)
        endRunID = in_dict.get("EndRun", None)
        tcks = in_dict.get("TCK")
        jobStart = in_dict.get("JobStartDate", None)
        jobEnd = in_dict.get("JobEndDate", None)
        smog2States = in_dict.get("SMOG2", None)
        extendedDQOK = in_dict.get("ExtendedDQOK", None)

        method = cls.bkkDB.getFilesWithMetadata
        force_kwargs = {}
        if parameters := in_dict.get("OnlyParameters", None):
            force_kwargs["parameters"] = parameters
        if seed_md5 := in_dict.get("SampleSeedMD5", None):
            force_kwargs["seed_md5"] = seed_md5
        if sample_max := in_dict.get("SampleMax", None):
            force_kwargs["sample_max"] = sample_max

        if force_kwargs:
            # Only the new implementation supports this
            method = partial(cls.bkkDB._newdb.getFilesWithMetadata, **force_kwargs)
        else:
            # TODO: Can be replaced with GETFILESWITHMETADATA_NAME_TO_COL
            parameters = [
                "FileName",
                "EventStat",
                "FileSize",
                "CreationDate",
                "JobStart",
                "JobEnd",
                "WorkerNode",
                "FileType",
                "RunNumber",
                "FillNumber",
                "FullStat",
                "DataqualityFlag",
                "EventInputStat",
                "TotalLuminosity",
                "Luminosity",
                "InstLuminosity",
                "TCK",
                "GUID",
                "ADLER32",
                "EventType",
                "MD5SUM",
                "VisibilityFlag",
                "JobId",
                "GotReplica",
                "InsertTimeStamp",
            ]

        if "EventTypeId" in in_dict:
            cls.log.verbose("The EventTypeId has to be replaced by EventType!")

        if "Quality" in in_dict:
            cls.log.verbose("The Quality has to be replaced by DataQuality!")

        retVal = method(
            configName,
            configVersion,
            conddescription,
            processing,
            evt,
            production,
            filetype,
            quality,
            visible,
            replicaFlag,
            startDate,
            endDate,
            runnumbers,
            startRunID,
            endRunID,
            tcks,
            jobStart,
            jobEnd,
            None,
            smog2States,
            extendedDQOK,
        )
        if not retVal["OK"]:
            return retVal
        records = []
        for record in retVal["Value"]:
            records += [list(record)]
        return S_OK({"ParameterNames": parameters, "Records": records, "TotalRecords": len(records)})

    #############################################################################
    types_getFilesSummary = [dict]

    @checkEnoughBKArguments
    def export_getFilesSummary(self, in_dict):
        """It returns sumary for a given data set.

        Input parameter is a dictionary which has the following keys:
        'ConfigName', 'ConfigVersion', 'ConditionDescription', 'EventType',
        'ProcessingPass','Production','RunNumber', 'FileType', DataQuality
        """
        self.log.debug("Input:", f"{in_dict}")

        configName = in_dict.get("ConfigName", default)
        configVersion = in_dict.get("ConfigVersion", default)
        condDescription = in_dict.get(
            "ConditionDescription",  # old single
            in_dict.get("SimulationConditions", in_dict.get("DataTakingConditions", default)),
        )
        processingPass = in_dict.get("ProcessingPass", default)
        eventType = in_dict.get("EventType", in_dict.get("EventTypeId", default))
        production = in_dict.get("Production", default)
        fileType = in_dict.get("FileType", default)
        dataQuality = in_dict.get("DataQuality", in_dict.get("DataQualityFlag", in_dict.get("Quality", default)))
        startRun = in_dict.get("StartRun", None)
        endRun = in_dict.get("EndRun", None)
        visible = in_dict.get("Visible", "Y")
        startDate = in_dict.get("StartDate", None)
        endDate = in_dict.get("EndDate", None)
        runNumbers = in_dict.get("RunNumber", in_dict.get("RunNumbers", []))
        replicaFlag = in_dict.get("ReplicaFlag", "Yes")
        tcks = in_dict.get("TCK")
        jobStart = in_dict.get("JobStartDate", None)
        jobEnd = in_dict.get("JobEndDate", None)
        smog2States = in_dict.get("SMOG2", None)
        extendedDQOK = in_dict.get("ExtendedDQOK", None)

        if "EventTypeId" in in_dict:
            self.log.verbose("The EventTypeId has to be replaced by EventType!")

        if "Quality" in in_dict:
            self.log.verbose("The Quality has to be replaced by DataQuality!")

        retVal = self.bkkDB.getFilesSummary(
            configName=configName,
            configVersion=configVersion,
            conditionDescription=condDescription,
            processingPass=processingPass,
            eventType=eventType,
            production=production,
            fileType=fileType,
            dataQuality=dataQuality,
            startRun=startRun,
            endRun=endRun,
            visible=visible,
            startDate=startDate,
            endDate=endDate,
            runNumbers=runNumbers,
            replicaFlag=replicaFlag,
            tcks=tcks,
            jobStart=jobStart,
            jobEnd=jobEnd,
            smog2States=smog2States,
            dqok=extendedDQOK,
        )
        if not retVal["OK"]:
            return retVal
        records = []
        parameters = ["NbofFiles", "NumberOfEvents", "FileSize", "Luminosity", "InstLuminosity"]
        for record in retVal["Value"]:
            records += [[record[0], record[1], record[2], record[3], record[4]]]
        return S_OK({"ParameterNames": parameters, "Records": records, "TotalRecords": len(records)})

    #############################################################################
    types_getLimitedFiles = [dict]

    @classmethod
    def export_getLimitedFiles(cls, in_dict):
        """It returns a chunk of files.

        This method is equivalent to the getFiles.
        """
        configName = in_dict.get("ConfigName", default)
        configVersion = in_dict.get("ConfigVersion", default)
        conddescription = in_dict.get(
            "ConditionDescription",  # old single
            in_dict.get("SimulationConditions", in_dict.get("DataTakingConditions", default)),
        )
        processing = in_dict.get("ProcessingPass", default)
        evt = in_dict.get("EventType", in_dict.get("EventTypeId", default))
        production = in_dict.get("Production", default)
        filetype = in_dict.get("FileType", default)
        quality = in_dict.get("DataQuality", in_dict.get("DataQualityFlag", in_dict.get("Quality", default)))
        runnb = in_dict.get("RunNumbers", in_dict.get("RunNumber", default))
        start = in_dict.get("StartItem", 0)
        maxValue = in_dict.get("MaxItem", 10)

        if "EventTypeId" in in_dict:
            cls.log.verbose("The EventTypeId has to be replaced by EventType!")

        if "Quality" in in_dict:
            cls.log.verbose("The Quality has to be replaced by DataQuality!")

        retVal = cls.bkkDB.getLimitedFiles(
            configName,
            configVersion,
            conddescription,
            processing,
            evt,
            production,
            filetype,
            quality,
            runnb,
            start,
            maxValue,
        )
        if not retVal["OK"]:
            return retVal
        records = []
        parameters = [
            "Name",
            "EventStat",
            "FileSize",
            "CreationDate",
            "JobStart",
            "JobEnd",
            "WorkerNode",
            "FileType",
            "EventType",
            "RunNumber",
            "FillNumber",
            "FullStat",
            "DataqualityFlag",
            "EventInputStat",
            "TotalLuminosity",
            "Luminosity",
            "InstLuminosity",
            "TCK",
            "WNMJFHS06",
            "HLT2TCK",
            "NumberOfProcessors",
        ]
        for record in retVal["Value"]:
            records += [
                [
                    record[0],
                    record[1],
                    record[2],
                    str(record[3]),
                    str(record[4]),
                    str(record[5]),
                    record[6],
                    record[7],
                    record[8],
                    record[9],
                    record[10],
                    record[11],
                    record[12],
                    record[13],
                    record[14],
                    record[15],
                    record[16],
                    record[17],
                ]
            ]
        return S_OK({"ParameterNames": parameters, "Records": records, "TotalRecords": len(records)})

    #############################################################################
    types_getAvailableDataQuality = []

    @classmethod
    def export_getAvailableDataQuality(cls):
        """it returns all the available data quality flags."""
        return cls.bkkDB.getAvailableDataQuality()

    #############################################################################
    types_getAvailableExtendedDQOK = []

    @classmethod
    def export_getAvailableExtendedDQOK(cls):
        """it returns all available Exdended DQOK flags."""
        return cls.bkkDB.getAvailableExtendedDQOK()

    #############################################################################
    types_getAvailableSMOG2States = []

    @classmethod
    def export_getAvailableSMOG2States(cls):
        """it returns all available SMOG2 states."""
        return cls.bkkDB.getAvailableSMOG2States()

    #############################################################################
    types_getAvailableProductions = []

    @classmethod
    def export_getAvailableProductions(cls):
        """It returns all the available productions which have associated file with
        replica flag yes."""
        return cls.bkkDB.getAvailableProductions()

    #############################################################################
    types_getAvailableRuns = []

    @classmethod
    def export_getAvailableRuns(cls):
        """It returns all the available runs which have associated files with
        reploica flag yes."""
        return cls.bkkDB.getAvailableRuns()

    #############################################################################
    types_getAvailableEventTypes = []

    @classmethod
    def export_getAvailableEventTypes(cls):
        """It returns all the available event types."""
        return cls.bkkDB.getAvailableEventTypes()

    #############################################################################
    types_getMoreProductionInformations = [int]

    @classmethod
    def export_getMoreProductionInformations(cls, prodid):
        """It returns inforation about a production."""
        return cls.bkkDB.getMoreProductionInformations(prodid)

    #############################################################################
    types_getJobInfo = [str]

    @classmethod
    def export_getJobInfo(cls, lfn):
        """It returns the job metadata information for a given lfn produced by this
        job."""
        return cls.bkkDB.getJobInfo(lfn)

    #############################################################################
    types_bulkJobInfo = [dict]

    @classmethod
    def export_bulkJobInfo(cls, lfns):
        """more info in the BookkeepingClient.py."""
        return cls.bkkDB.bulkJobInfo(lfns)

    #############################################################################
    types_getJobInformation = [dict]

    @classmethod
    def export_getJobInformation(cls, in_dict):
        """It returns the job metadata information for a given lfn produced by this
        job."""
        return cls.bkkDB.getJobInformation(in_dict)

    #############################################################################
    types_getRunNumber = [str]

    @classmethod
    def export_getRunNumber(cls, lfn):
        """It returns the run number for a given lfn!"""
        return cls.bkkDB.getRunNumber(lfn)

    #############################################################################
    types_getRunNbAndTck = [str]

    @classmethod
    def export_getRunNbAndTck(cls, lfn):
        """It returns the run number and tck for a given LFN."""
        return cls.bkkDB.getRunNbAndTck(lfn)

    #############################################################################
    types_getProductionFiles = [int, str]

    @classmethod
    def export_getProductionFiles(cls, prod, fileType, replica=default):
        """It returns files and their metadata for a given production, file type
        and replica."""
        return cls.bkkDB.getProductionFiles(prod, fileType, replica)

    #############################################################################
    types_getProductionFilesBulk = [list, (str, list)]

    @classmethod
    def export_getProductionFilesBulk(cls, prods, fileType, replica=default):
        """It returns files and their metadata for list of productions, file type
        and replica."""
        return cls.bkkDB.getProductionFilesBulk(prods, fileType, replica)

    #############################################################################
    types_getAvailableRunNumbers = []

    def export_getAvailableRunNumbers(self):
        """more info in the BookkeepingClient.py."""
        return self.export_getAvailableRuns()

    #############################################################################
    types_getRunFiles = [int]

    @classmethod
    def export_getRunFiles(cls, runid):
        """It returns all the files and their metadata for a given run number!"""
        return cls.bkkDB.getRunFiles(runid)

    #############################################################################
    types_updateFileMetaData = [str, dict]

    @classmethod
    def export_updateFileMetaData(cls, filename, fileAttr):
        """This method used to modify files metadata.

        Input parametes is a stirng (filename) and a dictionary (fileAttr)
        with the file attributes. {'GUID':34826386286382,'EventStat':222222}
        """
        return cls.bkkDB.updateFileMetaData(filename, fileAttr)

    #############################################################################
    types_renameFile = [str, str]

    @classmethod
    def export_renameFile(cls, oldLFN, newLFN):
        """It allows to change the name of a file which is in the Bookkeeping
        Metadata Catalogue."""
        return cls.bkkDB.renameFile(oldLFN, newLFN)

    #############################################################################
    types_getProductionProcessingPassID = [int]

    @classmethod
    def export_getProductionProcessingPassID(cls, prodid):
        """more info in the BookkeepingClient.py."""
        return cls.bkkDB.getProductionProcessingPassID(prodid)

    #############################################################################
    types_getProductionProcessingPass = [int]

    @classmethod
    def export_getProductionProcessingPass(cls, prodid):
        """It returns the processing pass for a given production."""
        return cls.bkkDB.getProductionProcessingPass(prodid)

    #############################################################################
    types_insertTag = [dict]

    @classmethod
    def export_insertTag(cls, values):
        """It used to register tags (CONDB, DDDB, etc) to the database.

        The input parameter is dictionary: {'TagName':'Value'}
        """
        successful = {}
        failed = {}

        for i in values:
            tags = values[i]
            for tag in tags:
                retVal = cls.bkkDB.existsTag(i, tag)
                if retVal["OK"] and not retVal["Value"]:
                    retVal = cls.bkkDB.insertTag(i, tag)
                    if not retVal["OK"]:
                        failed[tag] = i
                    else:
                        successful[tag] = i
                else:
                    failed[tag] = i
        return S_OK({"Successful": successful, "Failed": failed})

    #############################################################################
    types_setFileDataQuality = [list, str]

    @classmethod
    def export_setFileDataQuality(cls, lfns, flag):
        """more info in the BookkeepingClient.py."""
        return cls.bkkDB.setFileDataQuality(lfns, flag)

    #############################################################################
    types_setRunAndProcessingPassDataQuality = [int, str, str]

    @classmethod
    def export_setRunAndProcessingPassDataQuality(cls, runNB, procpass, flag):
        """It sets the data quality to a run which belong to a given processing
        pass.

        This method insert a new row to the runquality table. This value
        used to set the data quality flag to a given run files which
        processed by a given processing pass.
        """
        return cls.bkkDB.setRunAndProcessingPassDataQuality(runNB, procpass, flag)

    #############################################################################
    types_setRunDataQuality = [int, str]

    @classmethod
    def export_setRunDataQuality(cls, runNb, flag):
        """It sets the data quality for a given run!

        The input parameter is the run number and a data quality flag.
        """
        return cls.bkkDB.setRunDataQuality(runNb, flag)

    #############################################################################
    types_setExtendedDQOK = [int, bool, list]

    @classmethod
    def export_setExtendedDQOK(cls, runNB, update, dqok):
        """For given run insert or update the list of Systems which are OK
        NOTE: this call is priviledged
        """
        return cls.bkkDB.setExtendedDQOK(runNB, update, dqok)

    #############################################################################
    types_setProductionDataQuality = [int, str]

    @classmethod
    def export_setProductionDataQuality(cls, prod, flag):
        """It sets the data quality for a given production!"""
        return cls.bkkDB.setProductionDataQuality(prod, flag)

    #############################################################################
    types_getFileAncestors = [list, int, bool]

    @classmethod
    def export_getFileAncestors(cls, lfns, depth=None, replica=None):
        """more info in the BookkeepingClient.py."""
        return cls.bkkDB.getFileAncestors(lfns, depth, replica)

    #############################################################################
    types_getAncestors = [list, int]

    @classmethod
    def export_getAncestors(cls, lfns, depth):
        """Get the ancestors for a list of LFNs in input"""
        retVal = cls.bkkDB.getFileAncestors(lfns, depth, True)
        if not retVal["OK"]:
            return retVal
        values = retVal["Value"]
        for key, value in values["Successful"].items():
            values["Successful"][key] = [i["FileName"] for i in value]
        return S_OK(values)

    #############################################################################
    types_getFileDescendents = [list, int, int, bool]

    @classmethod
    def export_getFileDescendents(cls, lfn, depth, production=0, checkreplica=True):
        """more info in the BookkeepingClient.py."""
        return cls.bkkDB.getFileDescendents(lfn, depth, production, checkreplica)

    #############################################################################
    types_getFileDescendants = [list, int, int, bool]

    @deprecated("Use getFileDescendents")
    def export_getFileDescendants(self, lfn, depth, production=0, checkreplica=True):
        """more info in the BookkeepingClient.py."""
        return self.export_getFileDescendents(lfn, depth, production, checkreplica)

    #############################################################################
    types_checkfile = [str]

    @classmethod
    def export_checkfile(cls, fileName):
        """more info in the BookkeepingClient.py."""
        return cls.bkkDB.checkfile(fileName)

    #############################################################################
    types_checkFileTypeAndVersion = [str, str]

    @classmethod
    def export_checkFileTypeAndVersion(cls, ftype, version):
        """more info in the BookkeepingClient.py."""
        return cls.bkkDB.checkFileTypeAndVersion(ftype, version)

    #############################################################################
    types_checkEventType = [int]

    @classmethod
    def export_checkEventType(cls, eventTypeId):
        """more info in the BookkeepingClient.py."""
        if eventTypeId not in cls.__eventTypeCache:
            retVal = cls.bkkDB.checkEventType(eventTypeId)
            if not retVal["OK"]:
                return retVal
            cls.__eventTypeCache[eventTypeId] = retVal

        return cls.__eventTypeCache[eventTypeId]

    #############################################################################
    types_insertSimConditions = [dict]

    @classmethod
    def export_insertSimConditions(cls, in_dict):
        """It inserts a simulation condition to the Bookkeeping Metadata
        catalogue."""
        return cls.bkkDB.insertSimConditions(in_dict)

    #############################################################################
    types_getSimConditions = []

    @classmethod
    def export_getSimConditions(cls):
        """It returns all the simulation conditions which are in the Bookkeeping
        Metadata catalogue."""
        return cls.bkkDB.getSimConditions()

    #############################################################################
    types_removeReplica = [str]

    @classmethod
    def export_removeReplica(cls, fileName):
        """more info in the BookkeepingClient.py."""
        return cls.bkkDB.removeReplica(fileName)

    #############################################################################
    types_getFileMetadata = [list]

    @classmethod
    def export_getFileMetadata(cls, lfns):
        """more info in the BookkeepingClient.py."""
        return cls.bkkDB.getFileMetadata(lfns)

    #############################################################################
    types_getProductionFilesForWeb = [int, dict, dict, int, int]

    @classmethod
    def export_getProductionFilesWeb(cls, prod, ftype, sortDict, startItem, maxitems):
        """It returns files and their metadata information for a given
        production."""
        return cls.bkkDB.getProductionFilesForWeb(prod, ftype, sortDict, startItem, maxitems)

    #############################################################################
    types_exists = [list]

    @classmethod
    def export_exists(cls, lfns):
        """more info in the BookkeepingClient.py."""
        return cls.bkkDB.exists(lfns)

    #############################################################################
    types_addReplica = [list]

    @classmethod
    def export_addReplica(cls, fileName):
        """more info in the BookkeepingClient.py."""
        return cls.bkkDB.addReplica(fileName)

    #############################################################################
    types_getRunInformations = [int]

    @classmethod
    def export_getRunInformations(cls, runnb):
        """It returns run information and statistics."""
        return cls.bkkDB.getRunInformations(runnb)

    #############################################################################
    types_getRunInformation = [dict]

    @classmethod
    def export_getRunInformation(cls, runnb):
        """more info in the BookkeepingClient.py."""
        return cls.bkkDB.getRunInformation(runnb)

    #############################################################################
    types_getFileCreationLog = [str]

    @classmethod
    def export_getFileCreationLog(cls, lfn):
        """For a given file returns the log files of the job which created it."""
        return cls.bkkDB.getFileCreationLog(lfn)

    #############################################################################
    types_insertEventType = [int, str, str]

    @classmethod
    def export_insertEventType(cls, evid, desc, primary):
        """It inserts an event type to the Bookkeeping Metadata catalogue."""
        retVal = cls.bkkDB.checkEventType(evid)
        if not retVal["OK"]:  # meaning the event type is not already inserted
            retVal = cls.bkkDB.insertEventTypes(evid, desc, primary)
            if not retVal["OK"]:
                return retVal
            return S_OK(str(evid) + " event type added successfully!")
        return S_OK(str(evid) + " event type exists")

    #############################################################################
    types_updateEventType = [int, str, str]

    @classmethod
    def export_updateEventType(cls, evid, desc, primary):
        """It can used to modify an existing event type."""

        retVal = cls.bkkDB.checkEventType(evid)
        if not retVal["OK"]:
            return S_ERROR(str(evid) + " event type is missing in the BKK database!")

        retVal = cls.bkkDB.updateEventType(evid, desc, primary)
        if not retVal["OK"]:
            return retVal
        return S_OK(str(evid) + " event type updated successfully!")

    #############################################################################
    types_addFiles = [list]

    @classmethod
    def export_addFiles(cls, lfns):
        """more info in the BookkeepingClient.py."""
        return cls.bkkDB.addReplica(lfns)

    #############################################################################
    types_removeFiles = [list]

    @classmethod
    def export_removeFiles(cls, lfns):
        """more info in the BookkeepingClient.py."""
        return cls.bkkDB.removeReplica(lfns)

    #############################################################################
    types_getProductionSummary = [dict]

    @classmethod
    def export_getProductionSummary(cls, in_dict):
        """It can used to count the number of events for a given dataset."""

        cName = in_dict.get("ConfigName", default)
        cVersion = in_dict.get("ConfigVersion", default)
        production = in_dict.get("Production", default)
        simdesc = in_dict.get(
            "ConditionDescription",  # old single
            in_dict.get("SimulationConditions", in_dict.get("DataTakingConditions", default)),
        )
        pgroup = in_dict.get("ProcessingPass", default)
        ftype = in_dict.get("FileType", default)
        evttype = in_dict.get("EventType", default)
        return cls.bkkDB.getProductionSummary(cName, cVersion, simdesc, pgroup, production, ftype, evttype)

    #############################################################################
    types_getProductionInformation = [int]

    def export_getProductionInformation(self, prodid):
        """It returns statistics (data processing phases, number of events, etc.) for a given production.

        NOTE: This method is EXTREMELY slow and should be used with caution.
        NOTE: The returned paths are incorrect if we have no files GotReplica=True
        """

        nbjobs = None
        nbOfFiles = None
        nbOfEvents = None
        prodinfos = None

        value = self.bkkDB.getProductionNbOfJobs(prodid)
        if value["OK"]:
            nbjobs = value["Value"]

        value = self.bkkDB.getProductionNbOfFiles(prodid)
        if value["OK"]:
            nbOfFiles = value["Value"]

        value = self.bkkDB.getProductionNbOfEvents(prodid)
        if value["OK"]:
            nbOfEvents = value["Value"]

        value = self.bkkDB.getConfigsAndEvtType(prodid)
        if value["OK"]:
            prodinfos = value["Value"]

        path = "/"

        if not prodinfos:
            self.log.error("No Configs/Event type for production", prodid)
            return S_ERROR("No Configs/Event type")

        cname = prodinfos[0][0]
        cversion = prodinfos[0][1]
        path += cname + "/" + cversion + "/"

        res = self.bkkDB.getProductionSimulationCond(prodid)
        if not res["OK"]:
            return res
        path += res["Value"]

        res = self.bkkDB.getProductionProcessingPass(prodid)
        if not res["OK"]:
            return res
        path += res["Value"]
        prefix = "\n" + path

        # At this point prefix is of the form '\n/ConfigName/ConfigVersion/SimulationConditions/ProcessingPass'
        # We now iterate over nbOfEvents to get the paths containing the eventtype and filetype.
        # Each path is separated by a newline character.
        for filetype, _, eventtype, _ in nbOfEvents:
            path += prefix + "/" + str(eventtype) + "/" + filetype
        result = {
            "Production information": prodinfos,
            "Number of jobs": nbjobs,
            "Number of files": nbOfFiles,
            "Number of events": nbOfEvents,
            "Path": path,
        }
        return S_OK(result)

    #############################################################################
    types_getOutputPathsForProdID = [int]

    @convertToReturnValue
    def export_getOutputPathsForProdID(self, prodID):
        """It returns the file types for a given production."""

        configs = returnValueOrRaise(self.bkkDB.getConfigsAndEvtType(prodID))
        conddesc = returnValueOrRaise(self.bkkDB.getProductionSimulationCond(prodID))
        procpass = returnValueOrRaise(self.bkkDB.getProductionProcessingPass(prodID))
        filetypes = returnValueOrRaise(self.bkkDB.getFileTypesForProdID(prodID))
        paths = []
        for config_name, config_version, eventtype in configs:
            paths.extend(
                f"/{config_name}/{config_version}/{conddesc}{procpass}/{eventtype}/{filetype}" for filetype in filetypes
            )
        return paths

    #############################################################################
    types_getFileHistory = [str]

    @classmethod
    def export_getFileHistory(cls, lfn):
        """It returns all the information about a file."""
        retVal = cls.bkkDB.getFileHistory(lfn)
        result = {}
        records = []
        if retVal["OK"]:
            values = retVal["Value"]
            parameterNames = [
                "FileId",
                "FileName",
                "ADLER32",
                "CreationDate",
                "EventStat",
                "Eventtype",
                "Gotreplica",
                "GUI",
                "JobId",
                "md5sum",
                "FileSize",
                "FullStat",
                "Dataquality",
                "FileInsertDate",
                "Luminosity",
                "InstLuminosity",
            ]
            counter = 0
            for record in values:
                value = [
                    record[0],
                    record[1],
                    record[2],
                    record[3],
                    record[4],
                    record[5],
                    record[6],
                    record[7],
                    record[8],
                    record[9],
                    record[10],
                    record[11],
                    record[12],
                    record[13],
                    record[14],
                    record[15],
                ]
                records += [value]
                counter += 1
            result = {"ParameterNames": parameterNames, "Records": records, "TotalRecords": counter}
        else:
            result = S_ERROR(retVal["Message"])
        return S_OK(result)

    #############################################################################
    types_getJobsNb = [int]

    @classmethod
    @deprecated("Use getProductionNbOfJobs")
    def export_getJobsNb(cls, prodid):
        """more info in the BookkeepingClient.py."""
        return cls.bkkDB.getProductionNbOfJobs(prodid)

    #############################################################################
    types_getProductionNbOfJobs = [int]

    @classmethod
    def export_getProductionNbOfJobs(cls, prodid):
        """It returns the number of jobs for a given production."""
        return cls.bkkDB.getProductionNbOfJobs(prodid)

    #############################################################################
    types_getNumberOfEvents = [int]

    @classmethod
    @deprecated("Use getProductionNbOfEvents")
    def export_getNumberOfEvents(cls, prodid):
        """more info in the BookkeepingClient.py."""
        return cls.bkkDB.getProductionNbOfEvents(prodid)

    #############################################################################
    types_getProductionNbOfEvents = [int]

    @classmethod
    def export_getProductionNbOfEvents(cls, prodid):
        """It returns the number of events for a given production."""
        return cls.bkkDB.getProductionNbOfEvents(prodid)

    #############################################################################
    types_getSizeOfFiles = [int]

    @classmethod
    @deprecated("Use getProductionSizeOfFiles")
    def export_getSizeOfFiles(cls, prodid):
        """more info in the BookkeepingClient.py."""
        return cls.bkkDB.getProductionSizeOfFiles(prodid)

    #############################################################################
    types_getProductionSizeOfFiles = [int]

    @classmethod
    def export_getProductionSizeOfFiles(cls, prodid):
        """It returns the size of files for a given production."""
        return cls.bkkDB.getProductionSizeOfFiles(prodid)

    #############################################################################
    types_getNbOfFiles = [int]

    @classmethod
    @deprecated("Use getProductionNbOfFiles")
    def export_getNbOfFiles(cls, prodid):
        """more info in the BookkeepingClient.py."""
        return cls.bkkDB.getProductionNbOfFiles(prodid)

    #############################################################################
    types_getProductionNbOfFiles = [int]

    @classmethod
    def export_getProductionNbOfFiles(cls, prodid):
        """It returns the number of files produced by a given production."""
        return cls.bkkDB.getProductionNbOfFiles(prodid)

    #############################################################################
    types_getNbOfJobsBySites = [int]

    @classmethod
    def export_getNbOfJobsBySites(cls, prodid):
        """It returns the number of jobs executed at different sites for a given
        production."""
        return cls.bkkDB.getNbOfJobsBySites(prodid)

    #############################################################################
    types_getAvailableTags = []

    @classmethod
    def export_getAvailableTags(cls):
        """It returns the available database tags."""
        return cls.bkkDB.getAvailableTags()

    #############################################################################
    types_getProcessedEvents = [int]

    @classmethod
    @deprecated("Use getProductionProcessedEvents")
    def export_getProcessedEvents(cls, prodid):
        """more info in the BookkeepingClient.py."""
        return cls.bkkDB.getProductionProcessedEvents(prodid)

    #############################################################################
    types_getProductionProcessedEvents = [int]

    @classmethod
    def export_getProductionProcessedEvents(cls, prodid):
        """it returns the number of events processed for a given production."""
        cls.log.debug("getProductionProcessedEvents->Production:", "%d " % prodid)
        return cls.bkkDB.getProductionProcessedEvents(prodid)

    #############################################################################
    types_getRunsForAGivenPeriod = [dict]

    @classmethod
    def export_getRunsForAGivenPeriod(cls, in_dict):
        """It returns the available runs between a period.

        Input parameters:
        AllowOutsideRuns: If it is true, it only returns the runs which finished before EndDate.
        StartDate: the run start period
        EndDate: the run end period
        """
        return cls.bkkDB.getRunsForAGivenPeriod(in_dict)

    #############################################################################
    types_getProductionsFromView = [dict]

    @classmethod
    @deprecated("Useless?")
    def export_getProductionsFromView(cls, in_dict):
        """It returns the productions from the bookkeeping view for a given
        processing pass and run number.

        Input parameters: RunNumber ProcessingPass
        """
        # FIXME: might be a useless method
        return cls.bkkDB.getProductionsFromView(in_dict)

    #############################################################################
    types_getRunFilesDataQuality = [list]

    @classmethod
    def export_getRunFilesDataQuality(cls, runs):
        """more info in the BookkeepingClient.py."""
        return cls.bkkDB.getRunFilesDataQuality(runs)

    #############################################################################
    types_getRunExtendedDQOK = [int]

    @classmethod
    def export_getRunExtendedDQOK(cls, runNB):
        """Return the list of systems in ExtendedDQOK for given run"""
        return cls.bkkDB.getRunExtendedDQOK(runNB)

    #############################################################################
    types_setFilesInvisible = [list]

    @classmethod
    def export_setFilesInvisible(cls, lfns):
        """more info in the BookkeepingClient.py."""
        return cls.bkkDB.setFilesInvisible(lfns)

    #############################################################################
    types_setFilesVisible = [list]

    @classmethod
    def export_setFilesVisible(cls, lfns):
        """more info in the BookkeepingClient.py."""
        return cls.bkkDB.setFilesVisible(lfns)

    #############################################################################
    types_getRunAndProcessingPassDataQuality = [int, int]

    @classmethod
    def export_getRunAndProcessingPassDataQuality(cls, runnb, processing):
        """It returns the data quality flag for a given run and processing pass."""
        return cls.bkkDB.getRunAndProcessingPassDataQuality(runnb, processing)

    #############################################################################
    types_getAvailableConfigurations = []

    @classmethod
    def export_getAvailableConfigurations(cls):
        """It returns the available configurations."""
        return cls.bkkDB.getAvailableConfigurations()

    #############################################################################
    types_getRunProcessingPass = [int]

    @classmethod
    def export_getRunProcessingPass(cls, runnumber):
        """it returns the run number for a given run."""
        return cls.bkkDB.getRunProcessingPass(runnumber)

    #############################################################################
    types_getProductionFilesStatus = [int, list]

    @classmethod
    def export_getProductionFilesStatus(cls, productionid=None, lfns=None):
        """It returns the file status in the bkk for a given production or a list
        of lfns."""
        if not lfns:
            lfns = []
        return cls.bkkDB.getProductionFilesStatus(productionid, lfns)

    #############################################################################
    types_getFiles = [dict]

    @classmethod
    def export_getFiles(cls, values):
        """more info in the BookkeepingClient.py."""

        simdesc = default
        datataking = values.get(
            "SimulationConditions",  # old primary
            values.get("DataTakingConditions", values.get("ConditionDescription", default)),  # old secondary
        )
        procPass = values.get("ProcessingPass", default)
        ftype = values.get("FileType", default)
        evt = values.get("EventType", 0)
        configname = values.get("ConfigName", default)
        configversion = values.get("ConfigVersion", default)
        prod = values.get("Production", values.get("ProductionID", default))
        flag = values.get("DataQuality", values.get("DataQualityFlag", default))
        startd = values.get("StartDate", None)
        endd = values.get("EndDate", None)
        nbofevents = values.get("NbOfEvents", False)
        startRunID = values.get("StartRun", None)
        endRunID = values.get("EndRun", None)
        runNbs = values.get("RunNumber", values.get("RunNumbers", []))
        if not isinstance(runNbs, list):
            runNbs = [runNbs]
        replicaFlag = values.get("ReplicaFlag", "Yes")
        visible = values.get("Visible", default)
        filesize = values.get("FileSize", False)
        tck = values.get("TCK")
        jobStart = values.get("JobStartDate", None)
        jobEnd = values.get("JobEndDate", None)

        if "ProductionID" in values:
            cls.log.verbose("ProductionID will be removed. It will changed to Production")

        if "DataQualityFlag" in values:
            cls.log.verbose("DataQualityFlag will be removed. It will changed to DataQuality")

        if "RunNumbers" in values:
            cls.log.verbose("RunNumbers will be removed. It will changed to RunNumber")

        result = []
        retVal = cls.bkkDB.getFiles(
            simdesc,
            datataking,
            procPass,
            ftype,
            evt,
            configname,
            configversion,
            prod,
            flag,
            startd,
            endd,
            nbofevents,
            startRunID,
            endRunID,
            runNbs,
            replicaFlag,
            visible,
            filesize,
            tck,
            jobStart,
            jobEnd,
        )
        if not retVal["OK"]:
            return retVal
        values = retVal["Value"]
        for i in values:
            result += [i[0]]

        return S_OK(result)

    #############################################################################
    types_getFilesWithGivenDataSetsForUsers = [dict]

    def export_getFilesWithGivenDataSetsForUsers(self, values):
        """more info in the BookkeepingClient.py."""
        return self.export_getVisibleFilesWithMetadata(values)

    #############################################################################
    types_getVisibleFilesWithMetadata = [dict]

    @classmethod
    def export_getVisibleFilesWithMetadata(cls, in_dict):
        """It returns a list of files with metadata for a given condition."""

        conddescription = in_dict.get(
            "SimulationConditions",  # old primary
            in_dict.get("DataTakingConditions", in_dict.get("ConditionDescription", default)),  # old secondary
        )
        procPass = in_dict.get("ProcessingPass", default)
        ftype = in_dict.get("FileType", default)
        evt = in_dict.get("EventType", default)
        configname = in_dict.get("ConfigName", default)
        configversion = in_dict.get("ConfigVersion", default)
        prod = in_dict.get("Production", in_dict.get("ProductionID", default))
        dqflag = in_dict.get("DataQuality", in_dict.get("DataQualityFlag", default))
        startd = in_dict.get("StartDate", None)
        endd = in_dict.get("EndDate", None)
        startRunID = in_dict.get("StartRun", None)
        endRunID = in_dict.get("EndRun", None)
        runNbs = in_dict.get("RunNumber", in_dict.get("RunNumbers", []))
        replicaFlag = in_dict.get("ReplicaFlag", "Yes")
        tck = in_dict.get("TCK", [])
        visible = in_dict.get("Visible", "Y")
        jobStart = in_dict.get("JobStartDate", None)
        jobEnd = in_dict.get("JobEndDate", None)
        smog2States = in_dict.get("SMOG2", None)
        extendedDQOK = in_dict.get("ExtendedDQOK", None)

        if ftype == default:
            return S_ERROR("FileType is missing!")

        if "ProductionID" in in_dict:
            cls.log.verbose("ProductionID will be removed. It will changed to Production")

        if "DataQualityFlag" in in_dict:
            cls.log.verbose("DataQualityFlag will be removed. It will changed to DataQuality")

        if "RunNumbers" in in_dict:
            cls.log.verbose("RunNumbers will be removed. It will changed to RunNumber")

        cls.log.debug("getVisibleFilesWithMetadata->", str(in_dict))
        result = {}
        retVal = cls.bkkDB.getFilesWithMetadata(
            configName=configname,
            configVersion=configversion,
            conddescription=conddescription,
            processing=procPass,
            evt=evt,
            production=prod,
            filetype=ftype,
            quality=dqflag,
            visible=visible,
            replicaflag=replicaFlag,
            startDate=startd,
            endDate=endd,
            runnumbers=runNbs,
            startRunID=startRunID,
            endRunID=endRunID,
            tcks=tck,
            jobStart=jobStart,
            jobEnd=jobEnd,
            smog2States=smog2States,
            dqok=extendedDQOK,
        )

        summary = 0

        parameters = [
            "FileName",
            "EventStat",
            "FileSize",
            "CreationDate",
            "JobStart",
            "JobEnd",
            "WorkerNode",
            "FileType",
            "RunNumber",
            "FillNumber",
            "FullStat",
            "DataqualityFlag",
            "EventInputStat",
            "TotalLuminosity",
            "Luminosity",
            "InstLuminosity",
            "TCK",
            "GUID",
            "ADLER32",
            "EventType",
            "MD5SUM",
            "VisibilityFlag",
            "JobId",
            "GotReplica",
            "InsertTimeStamp",
        ]

        if not retVal["OK"]:
            return retVal

        values = retVal["Value"]
        nbfiles = 0
        nbevents = 0
        evinput = 0
        fsize = 0
        tLumi = 0
        lumi = 0
        ilumi = 0
        for i in values:
            nbfiles = nbfiles + 1
            row = dict(zip(parameters, i))
            if row["EventStat"] is not None:
                nbevents += row["EventStat"]
            if row["EventInputStat"] is not None:
                evinput += row["EventInputStat"]
            if row["FileSize"] is not None:
                fsize += row["FileSize"]
            if row["TotalLuminosity"] is not None:
                tLumi += row["TotalLuminosity"]
            if row["Luminosity"] is not None:
                lumi += row["Luminosity"]
            if row["InstLuminosity"] is not None:
                ilumi += row["InstLuminosity"]
            result[row["FileName"]] = {
                "EventStat": row["EventStat"],
                "EventInputStat": row["EventInputStat"],
                "Runnumber": row["RunNumber"],
                "Fillnumber": row["FillNumber"],
                "FileSize": row["FileSize"],
                "TotalLuminosity": row["TotalLuminosity"],
                "Luminosity": row["Luminosity"],
                "InstLuminosity": row["InstLuminosity"],
                "TCK": row["TCK"],
            }
        if nbfiles > 0:
            summary = {
                "Number Of Files": nbfiles,
                "Number of Events": nbevents,
                "EventInputStat": evinput,
                "FileSize": fsize / 1e9,
                "TotalLuminosity": tLumi,
                "Luminosity": lumi,
                "InstLuminosity": ilumi,
            }
        return S_OK({"LFNs": result, "Summary": summary})

    #############################################################################
    types_addProduction = [dict]

    @classmethod
    def export_addProduction(cls, infos):
        """It is used to register a production in the bkk.

        Input parameters:
        SimulationConditions
        DataTakingConditions
        Steps: the step which is used to process data for a given production.
        Production:
        InputProductionTotalProcessingPass: it is a path of the input data processing pass
        """

        cls.log.debug("Registering:", infos)
        simcond = infos.get("SimulationConditions", None)
        daqdesc = infos.get("DataTakingConditions", None)
        production = None

        if simcond is None and daqdesc is None:
            return S_ERROR("SimulationConditions and DataTakingConditions are both missing!")

        if "Steps" not in infos:
            return S_ERROR("Missing Steps!")
        if "Production" not in infos:
            return S_ERROR("Production is missing!")
        if "EventType" not in infos:
            return S_ERROR("EventType is missing!")

        steps = infos["Steps"]
        inputProdTotalProcessingPass = ""
        production = infos["Production"]
        inputProdTotalProcessingPass = infos.get("InputProductionTotalProcessingPass", "")
        configName = infos.get("ConfigName")
        configVersion = infos.get("ConfigVersion")
        eventType = infos.get("EventType")
        return cls.bkkDB.addProduction(
            production=production,
            simcond=simcond,
            daq=daqdesc,
            steps=steps,
            inputproc=inputProdTotalProcessingPass,
            configName=configName,
            configVersion=configVersion,
            eventType=eventType,
        )

    #############################################################################
    types_getEventTypes = [dict]

    @classmethod
    def export_getEventTypes(cls, in_dict):
        """It returns the available event types for a given configuration name and
        configuration version.

        Input parameters: ConfigName, ConfigVersion, Production
        """

        configName = in_dict.get("ConfigName", default)
        configVersion = in_dict.get("ConfigVersion", default)
        production = in_dict.get("Production", default)
        return cls.bkkDB.getEventTypes(configName, configVersion, production)

    #############################################################################
    types_getProcessingPassSteps = [dict]

    @classmethod
    def export_getProcessingPassSteps(cls, in_dict):
        """It returns the steps for a given stepname, processing pass and
        production."""
        stepname = in_dict.get("StepName", default)
        cond = in_dict.get(
            "ConditionDescription",  # old single
            in_dict.get("SimulationConditions", in_dict.get("DataTakingConditions", default)),
        )
        procpass = in_dict.get("ProcessingPass", default)

        return cls.bkkDB.getProcessingPassSteps(procpass, cond, stepname)

    #############################################################################
    types_getProductionProcessingPassSteps = [dict]

    @classmethod
    def export_getProductionProcessingPassSteps(cls, in_dict):
        """it returns the steps for a given production."""

        if "Production" in in_dict:
            return cls.bkkDB.getProductionProcessingPassSteps(in_dict["Production"])
        return S_ERROR("The Production dictionary key is missing!!!")

    #############################################################################
    types_getProductionOutputFiles = [dict]

    @deprecated("Use getProductionOutputFileTypes")
    def export_getProductionOutputFiles(self, in_dict):
        """more info in the BookkeepingClient.py."""

        return self.export_getProductionOutputFileTypes(in_dict)

    #############################################################################
    types_getProductionOutputFileTypes = [dict]

    @classmethod
    def export_getProductionOutputFileTypes(cls, in_dict):
        """It returns the output file types which produced by a given
        production."""

        production = in_dict.get("Production", default)
        stepid = in_dict.get("StepId", default)

        if production != default:
            return cls.bkkDB.getProductionOutputFileTypes(production, stepid)
        return S_ERROR("The Production dictionary key is missing!!!")

    #############################################################################
    types_getRunQuality = [str, str]

    @deprecated("Use getRunWithProcessingPassAndDataQuality")
    def export_getRunQuality(self, procpass, flag=default):
        """more info in the BookkeepingClient.py."""

        return self.export_getRunWithProcessingPassAndDataQuality(procpass, flag)

    #############################################################################
    types_getRunWithProcessingPassAndDataQuality = [str, str]

    @classmethod
    def export_getRunWithProcessingPassAndDataQuality(cls, procpass, flag=default):
        """It returns the run number for a given processing pass and a flag from
        the run quality table."""
        return cls.bkkDB.getRunWithProcessingPassAndDataQuality(procpass, flag)

    #############################################################################
    types_getRuns = [dict]

    @classmethod
    def export_getRuns(cls, in_dict):
        """It returns the runs for a given configuration name and version.

        Input parameters:
        """
        cName = in_dict.get("ConfigName", default)
        cVersion = in_dict.get("ConfigVersion", default)
        if cName != default and cVersion != default:
            return cls.bkkDB.getRuns(cName, cVersion)
        return S_ERROR("The configuration name and version have to be defined!")

    #############################################################################
    types_getRunProcPass = [dict]

    def export_getRunProcPass(self, in_dict):
        """more info in the BookkeepingClient.py."""
        return self.export_getRunAndProcessingPass(in_dict)

    #############################################################################
    types_getRunAndProcessingPass = [dict]

    @classmethod
    def export_getRunAndProcessingPass(cls, in_dict):
        """It returns all the processing pass and run number for a given run."""
        run = in_dict.get("RunNumber", default)
        if run != default:
            return cls.bkkDB.getRunAndProcessingPass(run)
        return S_ERROR("The run number has to be specified!")

    #############################################################################
    types_getProcessingPassId = [str]

    @classmethod
    def export_getProcessingPassId(cls, fullpath):
        """It returns the ProcessingPassId for a given path.

        this method should not used!
        """
        return cls.bkkDB.getProcessingPassId(fullpath)

    #############################################################################
    types_getRunNbFiles = [dict]

    @deprecated("Use getNbOfRawFiles")
    def export_getRunNbFiles(self, in_dict):
        """more info in the BookkeepingClient.py."""
        return self.export_getNbOfRawFiles(in_dict)

    #############################################################################
    types_getNbOfRawFiles = [dict]

    @classmethod
    def export_getNbOfRawFiles(cls, in_dict):
        """It counts the raw files for a given run and (or) event type."""

        runnb = in_dict.get("RunNumber", default)
        evt = in_dict.get("EventType", in_dict.get("EventTypeId", default))
        if "EventTypeId" in in_dict:
            cls.log.verbose("The EventTypeId has to be replaced by EventType!")

        replicaFlag = in_dict.get("ReplicaFlag", "Yes")
        visible = in_dict.get("Visible", "Y")
        isFinished = in_dict.get("Finished", "ALL")
        if runnb == default and evt == default:
            return S_ERROR("Run number or event type must be given!")
        retVal = cls.bkkDB.getNbOfRawFiles(runnb, evt, replicaFlag, visible, isFinished)
        if not retVal["OK"]:
            return retVal
        return S_OK(retVal["Value"][0][0])

    #############################################################################
    types_getTypeVersion = [list]

    def export_getTypeVersion(self, lfn):
        """more info in the BookkeepingClient.py."""
        return self.export_getFileTypeVersion(lfn)

    #############################################################################
    types_getFileTypeVersion = [list]

    @classmethod
    def export_getFileTypeVersion(cls, lfn):
        """more info in the BookkeepingClient.py."""
        return cls.bkkDB.getFileTypeVersion(lfn)

    #############################################################################
    types_getTCKs = [dict]

    @classmethod
    def export_getTCKs(cls, in_dict):
        """It returns the tcks for a given data set."""
        configName = in_dict.get("ConfigName", default)
        configVersion = in_dict.get("ConfigVersion", default)
        conddescription = in_dict.get(
            "ConditionDescription",  # old single
            in_dict.get("SimulationConditions", in_dict.get("DataTakingConditions", default)),
        )
        processing = in_dict.get("ProcessingPass", default)
        evt = in_dict.get("EventType", in_dict.get("EventTypeId", default))
        production = in_dict.get("Production", default)
        filetype = in_dict.get("FileType", default)
        quality = in_dict.get("DataQuality", in_dict.get("DataQualityFlag", in_dict.get("Quality", default)))
        runnb = in_dict.get("RunNumber", default)
        if "Quality" in in_dict:
            cls.log.verbose("The Quality has to be replaced by DataQuality!")

        if "EventTypeId" in in_dict:
            cls.log.verbose("The EventTypeId has to be replaced by EventType!")

        retVal = cls.bkkDB.getTCKs(
            configName, configVersion, conddescription, processing, evt, production, filetype, quality, runnb
        )
        if not retVal["OK"]:
            return retVal
        return S_OK([record[0] for record in retVal["Value"]])

    #############################################################################
    types_getAvailableTcks = [dict]

    def export_getAvailableTcks(self, in_dict):
        """more info in the BookkeepingClient.py."""
        return self.export_getTCKs(in_dict)

    #############################################################################
    types_getSteps = [int]

    @classmethod
    def export_getSteps(cls, prodID):
        """get list of steps used in a production"""
        return cls.bkkDB.getSteps(prodID)

    #############################################################################
    types_getStepsMetadata = [dict]

    @classmethod
    def export_getStepsMetadata(cls, in_dict):
        """It returns the step(s) which is produced  a given dataset."""
        cls.log.debug("getStepsMetadata", f"{in_dict}")
        configName = in_dict.get("ConfigName", default)
        configVersion = in_dict.get("ConfigVersion", default)
        cond = in_dict.get(
            "ConditionDescription",  # old single
            in_dict.get("SimulationConditions", in_dict.get("DataTakingConditions", default)),
        )
        procpass = in_dict.get("ProcessingPass", default)
        evt = in_dict.get("EventType", in_dict.get("EventTypeId", default))
        production = in_dict.get("Production", default)
        filetype = in_dict.get("FileType", default)
        runnb = in_dict.get("RunNumber", default)

        if "EventTypeId" in in_dict:
            cls.log.verbose("The EventTypeId has to be replaced by EventType!")

        if "Quality" in in_dict:
            cls.log.verbose("The Quality has to be replaced by DataQuality!")

        return cls.bkkDB.getStepsMetadata(configName, configVersion, cond, procpass, evt, production, filetype, runnb)

    #############################################################################
    types_getDirectoryMetadata_new = [list]

    @classmethod
    @deprecated("Use getDirectoryMetadata")
    def export_getDirectoryMetadata_new(cls, lfn):
        """more info in the BookkeepingClient.py."""
        return cls.bkkDB.getDirectoryMetadata(lfn)

        #############################################################################

    types_getDirectoryMetadata = [list]

    @classmethod
    def export_getDirectoryMetadata(cls, lfn):
        """more info in the BookkeepingClient.py."""
        cls.log.verbose("Getting the metadata for:", f"{lfn}")
        return cls.bkkDB.getDirectoryMetadata(lfn)

    #############################################################################
    types_getFilesForGUID = [str]

    @classmethod
    def export_getFilesForGUID(cls, guid):
        """It returns a file for a given GUID."""
        return cls.bkkDB.getFilesForGUID(guid)

    #############################################################################
    types_getRunsGroupedByDataTaking = []

    @classmethod
    def export_getRunsGroupedByDataTaking(cls):
        """It returns all the run numbers grouped by the data taking
        description."""
        return cls.bkkDB.getRunsGroupedByDataTaking()

    #############################################################################
    types_getListOfFills = [dict]

    @classmethod
    def export_getListOfFills(cls, in_dict):
        """It returns a list of FILL numbers for a given Configuration name,
        Configuration version and data taking description."""
        configName = in_dict.get("ConfigName", default)
        configVersion = in_dict.get("ConfigVersion", default)
        conddescription = in_dict.get(
            "ConditionDescription",  # old single
            in_dict.get("SimulationConditions", in_dict.get("DataTakingConditions", default)),
        )
        return cls.bkkDB.getListOfFills(configName, configVersion, conddescription)

    #############################################################################
    types_getRunsForFill = [int]

    @classmethod
    def export_getRunsForFill(cls, fillid):
        """It returns a list of runs for a given FILL."""
        return cls.bkkDB.getRunsForFill(fillid)

    #############################################################################
    types_getListOfRuns = [dict]

    @classmethod
    def export_getListOfRuns(cls, in_dict):
        """It returns a list of runs for a given conditions.

        Input parameter is a dictionary which has the following keys:
        'ConfigName', 'ConfigVersion', 'ConditionDescription',
        'EventType','ProcessingPass'
        """
        configName = in_dict.get("ConfigName", default)
        configVersion = in_dict.get("ConfigVersion", default)
        conddescription = in_dict.get(
            "ConditionDescription",  # old single
            in_dict.get("SimulationConditions", in_dict.get("DataTakingConditions", default)),
        )
        processing = in_dict.get("ProcessingPass", default)
        evt = in_dict.get("EventType", default)
        quality = in_dict.get("DataQuality", default)

        retVal = cls.bkkDB.getListOfRuns(configName, configVersion, conddescription, processing, evt, quality)
        if not retVal["OK"]:
            return retVal
        return S_OK([i[0] for i in retVal["Value"]])

    #############################################################################
    types_getListOfRunsInProd = [int]

    @classmethod
    def export_getListOfRunsInProd(cls, prod_id):
        """It returns a list of runs for a given production."""
        return cls.bkkDB.getListOfRunsInProd(prod_id)

    #############################################################################
    types_getInputOutputFilesForProd = [int, int]

    @classmethod
    def export_getInputOutputFilesForProd(cls, prod_id, run_number):
        """It returns a list of input and output files for a given production.

        The run number is used to filter the files as this results in faster queries.
        The list of available run numbers can be obtained using getListOfRunsInProd.
        """
        return cls.bkkDB.getInputOutputFilesForProd(prod_id, run_number)

    #################
    types_getOutputDescendantsForProd = [int, int]

    @classmethod
    def export_getOutputDescendantsForProd(cls, prod_id, run_number):
        """It returns the decendants of the output files for a given production.


        The run number is used to filter the files as this results in faster queries.
        The list of available run numbers can be obtained using getListOfRunsInProd.
        """
        return cls.bkkDB.getOutputDescendantsForProd(prod_id, run_number)

    #############################################################################
    types_getSimulationConditions = [dict]

    @classmethod
    def export_getSimulationConditions(cls, in_dict):
        """It returns a list of simulation conditions for a given conditions."""
        return cls.bkkDB.getSimulationConditions(in_dict)

    #############################################################################
    types_updateSimulationConditions = [dict]

    @classmethod
    def export_updateSimulationConditions(cls, in_dict):
        """It updates a given simulation condition."""
        return cls.bkkDB.updateSimulationConditions(in_dict)

    #############################################################################
    types_deleteSimulationConditions = [int]

    @classmethod
    def export_deleteSimulationConditions(cls, simid):
        """deletes a given simulation conditions."""
        return cls.bkkDB.deleteSimulationConditions(simid)

    #############################################################################
    types_getProductionSummaryFromView = [dict]

    @classmethod
    def export_getProductionSummaryFromView(cls, in_dict):
        """it returns a summary for a given condition."""
        return cls.bkkDB.getProductionSummaryFromView(in_dict)

    types_getJobInputOutputFiles = [list]

    @classmethod
    def export_getJobInputOutputFiles(cls, diracjobids):
        """It returns the input and output files for a given DIRAC jobid."""
        return cls.bkkDB.getJobInputOutputFiles(diracjobids)

    types_setRunOnlineFinished = [int]

    @classmethod
    def export_setRunOnlineFinished(cls, runnumber):
        """It is used to set the run finished..."""
        return cls.bkkDB.setRunStatusFinished(runnumber, "Y")

    types_setRunOnlineNotFinished = [int]

    @classmethod
    def export_setRunOnlineNotFinished(cls, runnumber):
        """You can set the runs not finished."""
        return cls.bkkDB.setRunStatusFinished(runnumber, "N")

    types_getRunStatus = [list]

    @classmethod
    def export_getRunStatus(cls, runnumbers):
        """it returns the status of the runs."""
        return cls.bkkDB.getRunStatus(runnumbers)

    types_bulkupdateFileMetaData = [dict]

    @classmethod
    def export_bulkupdateFileMetaData(cls, lfnswithmeta):
        """It updates the file metadata."""
        return cls.bkkDB.bulkupdateFileMetaData(lfnswithmeta)

    types_fixRunLuminosity = [list]

    @classmethod
    def export_fixRunLuminosity(cls, runnumbers):
        return cls.bkkDB.fixRunLuminosity(runnumbers)

    #############################################################################
    types_getProductionProducedEvents = [int]

    @classmethod
    def export_getProductionProducedEvents(cls, prodid):
        """it returns the number of events producced for a given production."""
        cls.log.debug("Retrieving the number of processed event for production", prodid)
        return cls.bkkDB.getProductionProducedEvents(prodid)

    #############################################################################
    types_bulkinsertEventType = [list]

    @classmethod
    def export_bulkinsertEventType(cls, eventtypes):
        """It inserts a list of event types to the db.

        :param eventtypes: it is a list of event types. For example, the list elements are the following

          .. code-block:: python

            {'EVTTYPEID': '12265021',
             'DESCRIPTION': 'Bu_D0pipipi,Kpi-withf2=DecProdCut_pCut1600MeV',
             'PRIMARY': '[B+ -> (D~0 -> K+ pi-) pi+ pi- pi+]cc'}


        :return: S_ERROR S_OK({'Failed':[],'Successful':[]})
        """
        return cls.bkkDB.bulkinsertEventType(eventtypes)

    #############################################################################
    types_bulkupdateEventType = [list]

    @classmethod
    def export_bulkupdateEventType(cls, eventtypes):
        """It updates a list of event types which are exist in the db.

        :param list eventtypes: it is a list of event types. For example: the list elements are the following:

          .. code-block:: python

          {'EVTTYPEID': '12265021',
          'DESCRIPTION': 'Bu_D0pipipi,Kpi-withf2=DecProdCut_pCut1600MeV',
          'PRIMARY': '[B+ -> (D~0 -> K+ pi-) pi+ pi- pi+]cc'}

        :return: S_ERROR S_OK({'Failed':[],'Successful':[]})
        """
        return cls.bkkDB.bulkupdateEventType(eventtypes)

    #############################################################################
    types_getRunConfigurationsAndDataTakingCondition = [int]

    @classmethod
    def export_getRunConfigurationsAndDataTakingCondition(cls, runnumber):
        """It returns minimal information for a given run.

        :param: int runnumber
        :return: S_OK()/S_ERROR ConfigName, ConfigVersion and DataTakingDescription
        """
        return cls.bkkDB.getRunConfigurationsAndDataTakingCondition(runnumber)

    types_deleteCertificationData = []

    @classmethod
    def export_deleteCertificationData(cls):
        """It destroy the data used by the integration test."""
        return cls.bkkDB.deleteCertificationData()

    types_updateProductionOutputfiles = []

    @classmethod
    def export_updateProductionOutputfiles(cls):
        """It is used to trigger an update of the productionoutputfiles table."""
        return cls.bkkDB.updateProductionOutputfiles()

    types_getAvailableTagsFromSteps = []

    @classmethod
    def export_getAvailableTagsFromSteps(cls):
        """It returns the all used datatbase tags: DDDB, CondDB, DQTag."""
        return cls.bkkDB.getAvailableTagsFromSteps()


class BookkeepingManagerHandler(BookkeepingManagerHandlerMixin, RequestHandler):
    pass
